""" Utility function used by the data hierarchy configuration module """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2023
#
#  Author: Jithu Murugan
#  Filename: utility_functions.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

from PySide6.QtCore import QEvent
from PySide6.QtGui import QMouseEvent
from PySide6.QtWidgets import QStyleOptionViewItem


def is_click_within_bounds(event: QEvent,
                           option: QStyleOptionViewItem) -> bool:
  """
  Check if the click event happened within the rect area of the QStyleOptionViewItem
  Args:
    event (QEvent): Mouse event captured from the view
    option (QStyleOptionViewItem): Option send during the edit event

  Returns (bool): True/False

  """
  if (event and option
      and event.type() == QEvent.Type.MouseButtonRelease
      and isinstance(event, QMouseEvent)):
    e = QMouseEvent(event)
    click_x = e.x()
    click_y = e.y()
    r = option.rect  # type: ignore[attr-defined]
    return (r.left() < click_x < r.left() + r.width()
            and r.top() < click_y < r.top() + r.height())
  return False



""" Long strings and dictionaries/JSON that would obfuscate code """
from typing import Any

defaultDocTypes: list[list[str]] = [
  #docType,               PURL,                                              title,          icon,                   shortcut, view
  ['x0',                  'http://purl.obolibrary.org/obo/NCIT_C47885',      'Projects',     '',                     'space', 'name,tags,.status,.objective,comment'],
  ['x1',                  'http://purl.obolibrary.org/obo/NCIT_C101129',     'Folders',      '',                      '',     ''],
  ['measurement',         'http://purl.obolibrary.org/obo/NCIT_C42790',      'Measurements', 'fa5s.thermometer-half', 'm',    'name,tags,comment,type,image,.sample,.workflow/procedure'],
  ['sample',              'http://purl.obolibrary.org/obo/NCIT_C19157',      'Samples',      'fa5s.vial',             's',    'name,tags,.chemistry,comment,qrCodes'],
  ['workflow',            'http://purl.obolibrary.org/obo/NCIT_C42753',      'Workflows',    'fa5s.list-ol',          'w',    'name,tags,comment'],
  ['workflow/procedure',  'https://schema.org/procedure',                    'Procedure',    '',                      '',     'name,tags,comment,content'],
  ['workflow/workplan' ,  'http://purl.obolibrary.org/obo/PROCO_0000093',    'Work plan',    '',                      '',     'name,tags,comment,content'],
  ['workflow/worklog'  ,  'http://dicom.nema.org/resources/ontology/DCM/LOG','Work log',     '',                      '',     'name,tags,comment,content'],
  ['instrument',          'http://purl.obolibrary.org/obo/NCIT_C16742',      'Instruments',  'ri.scales-2-line',      'i',    'name,tags,comment,.vendor'],
  ['instrument/extension','https://www.wikidata.org/wiki/Q19841649',         'Extensions',   '',                      '',     'name,tags,comment,.vendor'],
]


defaultSchema: list[list[str|int]] = [
  #docType,            group,index,key,        unit, mandatory, list
  ['x0',                  '', 0,   'name',      '',   'T',       ''],
  ['x0',                  '', 1,   'tags',      '',   '',        ''],
  ['x0',                  '', 2,   'status',    '',   '',        'active,paused,passive,finished'],
  ['x0',                  '', 3,   'objective', '',   '',        ''],
  ['x0',                  '', 4,   'comment',   '',   '',        ''],
  ['x1',                  '', 0,   'name',      '',   'T',       ''],
  ['x1',                  '', 1,   'tags',      '',   '',        ''],
  ['x1',                  '', 2,   'comment',   '',   '',        ''],
  ['measurement',         '', 0,   'name',      '',   'T',       ''],
  ['measurement',         '', 1,   'tags',      '',   '',        ''],
  ['measurement',         '', 2,   'comment',   '',   '',        ''],
  ['measurement',         '', 3,   'sample',    '',   '',        'sample'],
  ['measurement',         '', 4,   'workflow/procedure','','','workflow/procedure'],
  ['sample',              '', 0,   'name',      '',   'T',       ''],
  ['sample',              '', 1,   'tags',      '',   '',        ''],
  ['sample',              '', 2,   'chemistry', '',   '',        ''],
  ['sample',              '', 3,   'comment',   '',   '',        ''],
  ['sample',              '', 4,   'qrCodes',   '',   '',        ''],
  ['sample',      'geometry', 0,   'height',    'mm', '',        ''],
  ['sample',      'geometry', 1,   'width',     'mm', '',        ''],
  ['sample',      'geometry', 2,   'length',    'mm', '',        ''],
  ['workflow',            '', 0,   'name',      '',   'T',       ''],
  ['workflow',            '', 1,   'tags',      '',   '',        ''],
  ['workflow',            '', 2,   'comment',   '',   '',        ''],
  ['workflow',            '', 3,   'content',   '',   '',        ''],
  ['workflow/procedure',  '', 0,   'name',      '',   'T',       ''],
  ['workflow/procedure',  '', 1,   'tags',      '',   '',        ''],
  ['workflow/procedure',  '', 2,   'comment',   '',   '',        ''],
  ['workflow/procedure',  '', 3,   'content',   '',   '',        ''],
  ['workflow/workplan',   '', 0,   'name',      '',   'T',       ''],
  ['workflow/workplan',   '', 1,   'tags',      '',   '',        ''],
  ['workflow/workplan',   '', 2,   'comment',   '',   '',        ''],
  ['workflow/workplan',   '', 3,   'content',   '',   '',        ''],
  ['workflow/worklog',    '', 0,   'name',      '',   'T',       ''],
  ['workflow/worklog',    '', 1,   'tags',      '',   '',        ''],
  ['workflow/worklog',    '', 2,   'comment',   '',   '',        ''],
  ['workflow/worklog',    '', 3,   'content',   '',   '',        ''],
  ['instrument',          '', 0,   'name',      '',   'T',       ''],
  ['instrument',          '', 1,   'tags',      '',   '',        ''],
  ['instrument',          '', 2,   'comment',   '',   '',        ''],
  ['instrument',          '', 3,   'vendor',    '',   '',        ''],
  ['instrument/extension','', 0,   'name',      '',   'T',       ''],
  ['instrument/extension','', 1,   'tags',      '',   '',        ''],
  ['instrument/extension','', 2,   'comment',   '',   '',        ''],
  ['instrument/extension','', 3,   'vendor',    '',   '',        '']
]


defaultDefinitions = [
  ['name',              'What is the name this item?',                                    ''],
  ['tags',              'What are the tags?',                                             ''],
  ['status',            'What is the project status',                                     ''],
  ['objective',         'What is the objective?',                                         ''],
  ['comment',           'What are the comments?',                                         ''],
  ['content',           'What is procedure (Markdown possible; autofill if file given)?', ''],
  ['vendor',            'Who is the vendor?',                                             ''],
  ['sample',            'Which sample was used?',                                         ''],
  ['workflow/procedure','Which procedure was used?',                                      ''],
  ['chemistry',         'What is its chemical composition?',                              ''],
  ['qrCodes',           '',                           'https://www.wikidata.org/wiki/Q12203'],
  ['geometry.width',    'Sample width',                                                   ''],
  ['geometry.length',   'Sample length',                                                  ''],
  ['geometry.height',   'Sample height',                         'https://schema.org/height']
]


defaultDataHierarchyNode: list[dict[str, str]] = [
  {'docType':'-','class':'','idx':'0','name':'name',   'query':'What is the name?',     'unit':'','PURL':'','mandatory':'T','list':''},
  {'docType':'-','class':'','idx':'1','name':'tags',   'query':'What are the tags?',    'unit':'','PURL':'','mandatory':'','list':''},
  {'docType':'-','class':'','idx':'2','name':'comment','query':'What are the comments?','unit':'','PURL':'','mandatory':'','list':''}
  ]


CONF_FILE_NAME = '.pastaELN.json'

defaultConfiguration: dict[str, Any] = {
  'defaultProjectGroup': 'research',
  'userID': '$os.getlogin()$',
  'version': 0,
  'qrPrinter': {},
  'authors': [{'first': '', 'last': '', 'title': '', 'email': '', 'orcid': '',
               'organizations': [{'organization': '', 'rorid': ''}]}],
  'GUI': {},
  'projectGroups': {},
  'signingKeyPair': {},
  'repositories': {},
  'addOnParameter': {}
}

# level 1: type of property
#   within each: array of 3: description, default, all_choices
configurationGUI: dict[str, Any] = {
  'general': {
    'theme': ['Color style', 'none', ['amber', 'blue', 'cyan', 'pink', 'purple', 'teal', 'yellow', 'none']],
    'loggingLevel': ['Logging level (more->less)', 'INFO', ['DEBUG', 'INFO', 'WARNING', 'ERROR']],
    'autosave': ['Autosave entries in form', 'No', ['Yes', 'No']],
    'showProjectBtn': ['Show project button on top-left', 'Yes', ['Yes', 'No']],
    'checkForUpdates': ['Check for updates on startup', 'Yes', ['Yes', 'No']]
  },
  'dimensions': {
    'sidebarWidth': ['Sidebar width', 280, [220, 280, 340]],
    'maxTableColumnWidth': ['Maximum column width in tables', 400, [300, 400, 500, 600]],
    'imageSizeDetails': ['Image size in details view and form', 600, [300, 400, 500, 600]],
    'imageWidthProject': ['Image width in project view', 300, [200, 250, 300, 350, 400]],
    'maxProjectLeafHeight': ['Maximum height of item in project view', 250, [200, 250, 300, 400]],
    'widthContent': ['Width of procedures in project view', 600, [400, 500, 600, 700]],
    'docTypeOffset': ['Offset of document type in project view', 500, [400, 500, 600, 700]],
    'frameSize': ['Frame width around items in project view', 6, [4, 6, 8, 10]],
  }
}


DEFAULT_COLORS_PALETTE:dict[str,dict[str,str]] = {
    'dark': {
        'text': '#EEEEEE',
        'leafX': '#222222',
        'leafO': '#333333',
        'leafShadow': 'black',
    },
    'light': {
        'text': '#111111',
        'leafX': '#DDDDDD',
        'leafO': '#EEEEEE',
        'leafShadow': '#AAAAAA',
    },
}

SQLiteTranslationDict = {"'":'&prime;'}
SQLiteTranslation     = str.maketrans(SQLiteTranslationDict)

minimalDocInForm = {'tags':[], 'comment':'', '':{}}

SORTED_KEYS     = ['name', 'tags', 'comment', 'metaUser', '']
SORTED_DB_KEYS  = ['id', 'type', 'dateCreated','dateModified','dateSync', 'user','branch','gui','client','externalId','shasum']
DO_NOT_RENDER   = ['image','content','metaVendor','shasum','type','branch','gui','dateCreated',
                 'dateModified','id','user','name','externalId','client','elnIdentifier','oldIdentifier']

AboutMessage  = '### PASTA-ELN | The favorite ELN for experimental scientists\n\nPASTA-ELN provides a streamlined '\
                'and efficient solution for experimental scientists\n\n to manage and organize raw data alongside '\
                'associated metadata.\n\n'

setupText = """
### Welcome to the PASTA-ELN setup
Three components are needed for proper functioning of PASTA-ELN:
- Configuration of preferences
- Shortcut creation
- Example project creation

This setup will analyse and (possibly) correct these items.
If an attempt fails: please follow to this [website](https://pasta-eln.github.io/pasta-eln/install.html).
"""


exampleDataString = """
Do you want to create an example project?

This step helps to verify the installation and provides an helpful example for new users.

!WARNING! This process will RESET everything and thereby DELETE EVERYTHING since you installed pastaELN.

This step usually takes up to 20sec, so please be patient. Sometimes, Linux mentions that the program
is not responding and asks if to close/wait. Please WAIT.
"""

shortcuts = """
### Default shortcuts:

**Ctrl+Space**: List projects

**Ctrl+M**: List measurements

**Ctrl+S**: List samples

**Ctrl+W**: List workflows

**Ctrl+I**: List instruments

**Ctrl+T**: List tags

**Ctrl+U**: List unidentified

**F2**: Test file extraction

**F5**: Synchronize

**F9**: Restart

**Ctrl+?**: Verify database integrity

**Ctrl+0**: Configuration
"""


tableHeaderHelp = """
## You can add custom rows via bottom text area.

If you want to add a column:
- to add a normal data-field (comment, content, name, type, tags, user, date), enter this field : 'comment'
- to check the existence of an image: enter 'image'
- to check if a tag is present: "#tag", in which you replace "tag" by the tag you want to look for. "_curated" is a special tag for measurements.
- to get metadata, use a ".": e.g. "metaVendor.fileExtension", "metaUser.stress". Capitalization is important.
</ul>
"""

# FontAwesome 5 icons
allIcons = ['fa5s.address-book', 'fa5s.address-card', 'fa5s.adjust', 'fa5s.align-center', 'fa5s.align-justify',
            'fa5s.align-left', 'fa5s.align-right', 'fa5s.allergies', 'fa5s.ambulance',
            'fa5s.american-sign-language-interpreting', 'fa5s.anchor', 'fa5s.angle-double-down',
            'fa5s.angle-double-left', 'fa5s.angle-double-right', 'fa5s.angle-double-up', 'fa5s.angle-down',
            'fa5s.angle-left', 'fa5s.angle-right', 'fa5s.angle-up', 'fa5s.archive', 'fa5s.arrow-alt-circle-down',
            'fa5s.arrow-alt-circle-left', 'fa5s.arrow-alt-circle-right', 'fa5s.arrow-alt-circle-up',
            'fa5s.arrow-circle-down', 'fa5s.arrow-circle-left', 'fa5s.arrow-circle-right', 'fa5s.arrow-circle-up',
            'fa5s.arrow-down', 'fa5s.arrow-left', 'fa5s.arrow-right', 'fa5s.arrow-up', 'fa5s.arrows-alt',
            'fa5s.arrows-alt-h', 'fa5s.arrows-alt-v', 'fa5s.assistive-listening-systems', 'fa5s.asterisk', 'fa5s.at',
            'fa5s.audio-description', 'fa5s.backward', 'fa5s.balance-scale', 'fa5s.ban', 'fa5s.band-aid',
            'fa5s.barcode', 'fa5s.bars', 'fa5s.baseball-ball', 'fa5s.basketball-ball', 'fa5s.bath',
            'fa5s.battery-empty', 'fa5s.battery-full', 'fa5s.battery-half', 'fa5s.battery-quarter',
            'fa5s.battery-three-quarters', 'fa5s.bed', 'fa5s.beer', 'fa5s.bell', 'fa5s.bell-slash', 'fa5s.bicycle',
            'fa5s.binoculars', 'fa5s.birthday-cake', 'fa5s.blind', 'fa5s.bold', 'fa5s.bolt', 'fa5s.bomb', 'fa5s.book',
            'fa5s.bookmark', 'fa5s.bowling-ball', 'fa5s.box', 'fa5s.box-open', 'fa5s.boxes', 'fa5s.braille',
            'fa5s.briefcase', 'fa5s.briefcase-medical', 'fa5s.bug', 'fa5s.building', 'fa5s.bullhorn', 'fa5s.bullseye',
            'fa5s.burn', 'fa5s.bus', 'fa5s.calculator', 'fa5s.calendar', 'fa5s.calendar-alt', 'fa5s.calendar-check',
            'fa5s.calendar-minus', 'fa5s.calendar-plus', 'fa5s.calendar-times', 'fa5s.camera', 'fa5s.camera-retro',
            'fa5s.capsules', 'fa5s.car', 'fa5s.caret-down', 'fa5s.caret-left', 'fa5s.caret-right',
            'fa5s.caret-square-down', 'fa5s.caret-square-left', 'fa5s.caret-square-right', 'fa5s.caret-square-up',
            'fa5s.caret-up', 'fa5s.cart-arrow-down', 'fa5s.cart-plus', 'fa5s.certificate', 'fa5s.chart-area',
            'fa5s.chart-bar', 'fa5s.chart-line', 'fa5s.chart-pie', 'fa5s.check', 'fa5s.check-circle',
            'fa5s.check-square', 'fa5s.chess', 'fa5s.chess-bishop', 'fa5s.chess-board', 'fa5s.chess-king',
            'fa5s.chess-knight', 'fa5s.chess-pawn', 'fa5s.chess-queen', 'fa5s.chess-rook', 'fa5s.chevron-circle-down',
            'fa5s.chevron-circle-left', 'fa5s.chevron-circle-right', 'fa5s.chevron-circle-up', 'fa5s.chevron-down',
            'fa5s.chevron-left', 'fa5s.chevron-right', 'fa5s.chevron-up', 'fa5s.child', 'fa5s.circle',
            'fa5s.circle-notch', 'fa5s.clipboard', 'fa5s.clipboard-check', 'fa5s.clipboard-list', 'fa5s.clock',
            'fa5s.clone', 'fa5s.closed-captioning', 'fa5s.cloud', 'fa5s.cloud-download-alt', 'fa5s.cloud-upload-alt',
            'fa5s.code', 'fa5s.code-branch', 'fa5s.coffee', 'fa5s.cog', 'fa5s.cogs', 'fa5s.columns', 'fa5s.comment',
            'fa5s.comment-alt', 'fa5s.comment-dots', 'fa5s.comment-slash', 'fa5s.comments', 'fa5s.compass',
            'fa5s.compress', 'fa5s.copy', 'fa5s.copyright', 'fa5s.couch', 'fa5s.credit-card', 'fa5s.crop',
            'fa5s.crosshairs', 'fa5s.cube', 'fa5s.cubes', 'fa5s.cut', 'fa5s.database', 'fa5s.deaf', 'fa5s.desktop',
            'fa5s.diagnoses', 'fa5s.dna', 'fa5s.dollar-sign', 'fa5s.dolly', 'fa5s.dolly-flatbed', 'fa5s.donate',
            'fa5s.dot-circle', 'fa5s.dove', 'fa5s.download', 'fa5s.edit', 'fa5s.eject', 'fa5s.ellipsis-h',
            'fa5s.ellipsis-v', 'fa5s.envelope', 'fa5s.envelope-open', 'fa5s.envelope-square', 'fa5s.eraser',
            'fa5s.euro-sign', 'fa5s.exchange-alt', 'fa5s.exclamation', 'fa5s.exclamation-circle',
            'fa5s.exclamation-triangle', 'fa5s.expand', 'fa5s.expand-arrows-alt', 'fa5s.external-link-alt',
            'fa5s.external-link-square-alt', 'fa5s.eye', 'fa5s.eye-dropper', 'fa5s.eye-slash', 'fa5s.fast-backward',
            'fa5s.fast-forward', 'fa5s.fax', 'fa5s.female', 'fa5s.fighter-jet', 'fa5s.file', 'fa5s.file-alt',
            'fa5s.file-archive', 'fa5s.file-audio', 'fa5s.file-code', 'fa5s.file-excel', 'fa5s.file-image',
            'fa5s.file-medical', 'fa5s.file-medical-alt', 'fa5s.file-pdf', 'fa5s.file-powerpoint', 'fa5s.file-video',
            'fa5s.file-word', 'fa5s.film', 'fa5s.filter', 'fa5s.fire', 'fa5s.fire-extinguisher', 'fa5s.first-aid',
            'fa5s.flag', 'fa5s.flag-checkered', 'fa5s.flask', 'fa5s.folder', 'fa5s.folder-open', 'fa5s.font',
            'fa5s.football-ball', 'fa5s.forward', 'fa5s.frown', 'fa5s.futbol', 'fa5s.gamepad', 'fa5s.gavel', 'fa5s.gem',
            'fa5s.genderless', 'fa5s.gift', 'fa5s.glass-martini', 'fa5s.globe', 'fa5s.golf-ball', 'fa5s.graduation-cap',
            'fa5s.h-square', 'fa5s.hand-holding', 'fa5s.hand-holding-heart', 'fa5s.hand-holding-usd',
            'fa5s.hand-lizard', 'fa5s.hand-paper', 'fa5s.hand-peace', 'fa5s.hand-point-down', 'fa5s.hand-point-left',
            'fa5s.hand-point-right', 'fa5s.hand-point-up', 'fa5s.hand-pointer', 'fa5s.hand-rock', 'fa5s.hand-scissors',
            'fa5s.hand-spock', 'fa5s.hands', 'fa5s.hands-helping', 'fa5s.handshake', 'fa5s.hashtag', 'fa5s.hdd',
            'fa5s.heading', 'fa5s.headphones', 'fa5s.heart', 'fa5s.heartbeat', 'fa5s.history', 'fa5s.hockey-puck',
            'fa5s.home', 'fa5s.hospital', 'fa5s.hospital-alt', 'fa5s.hospital-symbol', 'fa5s.hourglass',
            'fa5s.hourglass-end', 'fa5s.hourglass-half', 'fa5s.hourglass-start', 'fa5s.i-cursor', 'fa5s.id-badge',
            'fa5s.id-card', 'fa5s.id-card-alt', 'fa5s.image', 'fa5s.images', 'fa5s.inbox', 'fa5s.indent',
            'fa5s.industry', 'fa5s.info', 'fa5s.info-circle', 'fa5s.italic', 'fa5s.key', 'fa5s.keyboard',
            'fa5s.language', 'fa5s.laptop', 'fa5s.leaf', 'fa5s.lemon', 'fa5s.level-down-alt', 'fa5s.level-up-alt',
            'fa5s.life-ring', 'fa5s.lightbulb', 'fa5s.link', 'fa5s.lira-sign', 'fa5s.list', 'fa5s.list-alt',
            'fa5s.list-ol', 'fa5s.list-ul', 'fa5s.location-arrow', 'fa5s.lock', 'fa5s.lock-open',
            'fa5s.long-arrow-alt-down', 'fa5s.long-arrow-alt-left', 'fa5s.long-arrow-alt-right',
            'fa5s.long-arrow-alt-up', 'fa5s.low-vision', 'fa5s.magic', 'fa5s.magnet', 'fa5s.male', 'fa5s.map',
            'fa5s.map-marker', 'fa5s.map-marker-alt', 'fa5s.map-pin', 'fa5s.map-signs', 'fa5s.mars', 'fa5s.mars-double',
            'fa5s.mars-stroke', 'fa5s.mars-stroke-h', 'fa5s.mars-stroke-v', 'fa5s.medkit', 'fa5s.meh', 'fa5s.mercury',
            'fa5s.microchip', 'fa5s.microphone', 'fa5s.microphone-slash', 'fa5s.minus', 'fa5s.minus-circle',
            'fa5s.minus-square', 'fa5s.mobile', 'fa5s.mobile-alt', 'fa5s.money-bill-alt', 'fa5s.moon',
            'fa5s.motorcycle', 'fa5s.mouse-pointer', 'fa5s.music', 'fa5s.neuter', 'fa5s.newspaper',
            'fa5s.notes-medical', 'fa5s.object-group', 'fa5s.object-ungroup', 'fa5s.outdent', 'fa5s.paint-brush',
            'fa5s.pallet', 'fa5s.paper-plane', 'fa5s.paperclip', 'fa5s.parachute-box', 'fa5s.paragraph', 'fa5s.paste',
            'fa5s.pause', 'fa5s.pause-circle', 'fa5s.paw', 'fa5s.pen-square', 'fa5s.pencil-alt', 'fa5s.people-carry',
            'fa5s.percent', 'fa5s.phone', 'fa5s.phone-slash', 'fa5s.phone-square', 'fa5s.phone-volume',
            'fa5s.piggy-bank', 'fa5s.pills', 'fa5s.plane', 'fa5s.play', 'fa5s.play-circle', 'fa5s.plug', 'fa5s.plus',
            'fa5s.plus-circle', 'fa5s.plus-square', 'fa5s.podcast', 'fa5s.poo', 'fa5s.pound-sign', 'fa5s.power-off',
            'fa5s.prescription-bottle', 'fa5s.prescription-bottle-alt', 'fa5s.print', 'fa5s.procedures',
            'fa5s.puzzle-piece', 'fa5s.qrcode', 'fa5s.question', 'fa5s.question-circle', 'fa5s.quidditch',
            'fa5s.quote-left', 'fa5s.quote-right', 'fa5s.random', 'fa5s.recycle', 'fa5s.redo', 'fa5s.redo-alt',
            'fa5s.registered', 'fa5s.reply', 'fa5s.reply-all', 'fa5s.retweet', 'fa5s.ribbon', 'fa5s.road',
            'fa5s.rocket', 'fa5s.rss', 'fa5s.rss-square', 'fa5s.ruble-sign', 'fa5s.rupee-sign', 'fa5s.save',
            'fa5s.search', 'fa5s.search-minus', 'fa5s.search-plus', 'fa5s.seedling', 'fa5s.server', 'fa5s.share',
            'fa5s.share-alt', 'fa5s.share-alt-square', 'fa5s.share-square', 'fa5s.shekel-sign', 'fa5s.shield-alt',
            'fa5s.ship', 'fa5s.shipping-fast', 'fa5s.shopping-bag', 'fa5s.shopping-basket', 'fa5s.shopping-cart',
            'fa5s.shower', 'fa5s.sign', 'fa5s.sign-in-alt', 'fa5s.sign-language', 'fa5s.sign-out-alt', 'fa5s.signal',
            'fa5s.sitemap', 'fa5s.sliders-h', 'fa5s.smile', 'fa5s.smoking', 'fa5s.snowflake', 'fa5s.sort',
            'fa5s.sort-alpha-down', 'fa5s.sort-alpha-up', 'fa5s.sort-amount-down', 'fa5s.sort-amount-up',
            'fa5s.sort-down', 'fa5s.sort-numeric-down', 'fa5s.sort-numeric-up', 'fa5s.sort-up', 'fa5s.space-shuttle',
            'fa5s.spinner', 'fa5s.square', 'fa5s.square-full', 'fa5s.star', 'fa5s.star-half', 'fa5s.step-backward',
            'fa5s.step-forward', 'fa5s.stethoscope', 'fa5s.sticky-note', 'fa5s.stop', 'fa5s.stop-circle',
            'fa5s.stopwatch', 'fa5s.street-view', 'fa5s.strikethrough', 'fa5s.subscript', 'fa5s.subway',
            'fa5s.suitcase', 'fa5s.sun', 'fa5s.superscript', 'fa5s.sync', 'fa5s.sync-alt', 'fa5s.syringe', 'fa5s.table',
            'fa5s.table-tennis', 'fa5s.tablet', 'fa5s.tablet-alt', 'fa5s.tablets', 'fa5s.tachometer-alt', 'fa5s.tag',
            'fa5s.tags', 'fa5s.tape', 'fa5s.tasks', 'fa5s.taxi', 'fa5s.terminal', 'fa5s.text-height', 'fa5s.text-width',
            'fa5s.th', 'fa5s.th-large', 'fa5s.th-list', 'fa5s.thermometer', 'fa5s.thermometer-empty',
            'fa5s.thermometer-full', 'fa5s.thermometer-half', 'fa5s.thermometer-quarter',
            'fa5s.thermometer-three-quarters', 'fa5s.thumbs-down', 'fa5s.thumbs-up', 'fa5s.thumbtack',
            'fa5s.ticket-alt', 'fa5s.times', 'fa5s.times-circle', 'fa5s.tint', 'fa5s.toggle-off', 'fa5s.toggle-on',
            'fa5s.trademark', 'fa5s.train', 'fa5s.transgender', 'fa5s.transgender-alt', 'fa5s.trash', 'fa5s.trash-alt',
            'fa5s.tree', 'fa5s.trophy', 'fa5s.truck', 'fa5s.truck-loading', 'fa5s.truck-moving', 'fa5s.tty', 'fa5s.tv',
            'fa5s.umbrella', 'fa5s.underline', 'fa5s.undo', 'fa5s.undo-alt', 'fa5s.universal-access', 'fa5s.university',
            'fa5s.unlink', 'fa5s.unlock', 'fa5s.unlock-alt', 'fa5s.upload', 'fa5s.user', 'fa5s.user-circle',
            'fa5s.user-md', 'fa5s.user-plus', 'fa5s.user-secret', 'fa5s.user-times', 'fa5s.users', 'fa5s.utensil-spoon',
            'fa5s.utensils', 'fa5s.venus', 'fa5s.venus-double', 'fa5s.venus-mars', 'fa5s.vial', 'fa5s.vials',
            'fa5s.video', 'fa5s.video-slash', 'fa5s.volleyball-ball', 'fa5s.volume-down', 'fa5s.volume-off',
            'fa5s.volume-up', 'fa5s.warehouse', 'fa5s.weight', 'fa5s.wheelchair', 'fa5s.wifi', 'fa5s.window-close',
            'fa5s.window-maximize', 'fa5s.window-minimize', 'fa5s.window-restore', 'fa5s.wine-glass', 'fa5s.won-sign',
            'fa5s.wrench', 'fa5s.x-ray', 'fa5s.yen-sign']



""" Python Backend: all operations with the filesystem are here """
import importlib
import json
import logging
import os
import sys
import tempfile
import traceback
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Callable, Optional, Union
from urllib import request
from .fixedStringsJson import CONF_FILE_NAME, configurationGUI, defaultConfiguration
from .miscTools import generic_hash
from .mixin_cli import CLI_Mixin
from .sqlite import SqlLiteDB
from .textTools.handleDictionaries import diffDicts, fillDocBeforeCreate
from .textTools.stringChanges import camelCase, createDirName, outputString


class Backend(CLI_Mixin):
  """
  PYTHON BACKEND
  """

  def __init__(self, defaultProjectGroup:str=''):
    """
    open server and define database

    Args:
        defaultProjectGroup (string): name of configuration / project-group used; if not given, use the one defined by 'defaultProjectGroup' in config file
    """
    #initialize basic values
    self.configFileName = Path.home() / CONF_FILE_NAME
    self.configuration: dict[str, Any] = {}
    self.hierStack:list[str] = []
    self.cwd:Optional[Path]  = Path('.')
    self.initialize(defaultProjectGroup)


  def initialize(self, defaultProjectGroup:str='') -> None:
    """
    initialize or reinitialize server and define database

    Args:
        defaultProjectGroup (string): name of configuration / project-group used; if not given, use the one defined by 'defaultProjectGroup' in config file
    """
    self.configuration = defaultConfiguration
    if self.configFileName.is_file():
      with open(self.configFileName, encoding='utf-8') as confFile:
        self.configuration |= json.load(confFile)
    for _, items in configurationGUI.items():
      for k,v in items.items():
        if k not in self.configuration['GUI']:
          self.configuration['GUI'][k] = v[1]
    if self.configuration['version'] != 3:
      print('**Info: configuration file does not exist or version is != 3')
      return
    defaultProjectGroup = defaultProjectGroup or self.configuration['defaultProjectGroup']
    if defaultProjectGroup not in self.configuration['projectGroups']:
      raise ValueError('BadConfigurationFileError')
    self.configurationProjectGroup = defaultProjectGroup
    projectGroup = self.configuration['projectGroups'][defaultProjectGroup]
    # directories
    #    self.basePath (root of directory tree) is root of all projects
    #    self.cwd changes during program but is similarly the full path from root
    self.basePath   = Path(projectGroup['local']['path'])
    self.cwd        = Path(projectGroup['local']['path'])
    self.addOnPath  = Path(projectGroup['addOnDir'])
    sys.path.append(str(self.addOnPath))  #allow add-ons
    # decipher miscellaneous configuration and store
    self.userID   = self.configuration['userID']
    # start database
    self.db = SqlLiteDB(basePath=self.basePath)
    # internal hierarchy structure
    self.hierStack = []
    return


  def exit(self) -> None:
    """
    Shutting down things
    """
    self.db.exit()
    return


  ######################################################
  ### Change in database
  ######################################################
  def editData(self, doc:dict[str,Any]) -> None:
    """
    Edit data from version 2 information flow by wrapping addData

    Args:
      doc (dict): dict to save
    """
    if doc['branch'][0]['path'] is None:
      self.cwd     = None
    else:
      self.cwd     = self.basePath/doc['branch'][0]['path']
    self.hierStack = doc['branch'][0]['stack']+[doc['id']]
    doc['childNum']= doc['branch'][0]['child']
    # change content
    doc = self.addData('-edit-', doc)
    if doc['id'].startswith('x'):
      pathStr = '' if self.cwd is None else str(self.cwd.relative_to(self.basePath))
      self.db.updateChildrenOfParentsChanges(pathStr, doc['branch'][0]['path'], '/'.join(self.hierStack),'')
    self.cwd = self.basePath #reset to sensible before continuing
    self.hierStack = []
    return


  def addData(self, docType:str, doc:dict[str,Any], hierStack:list[str]=[], localCopy:bool=False,
              forceNewImage:bool=False) -> dict[str,Any]:
    """
    Save doc to database, also after edit

    Args:
        docType (string): docType to be stored, subtypes are / separated; or '-edit-'
        doc (dict): to be stored
        hierStack (list): hierStack from external functions
        localCopy (bool): copy a remote file to local version
        forceNewImage (bool): create new image in any case

    Returns:
        str: docID, empty string if failure
    """
    doc['user']  = self.userID
    childNum     = doc.pop('childNum',None)
    path         = None
    oldPath      = None
    operation    = 'c'  #operation of branch/path
    if docType == '-edit-':
      edit = True
      if 'type' not in doc:
        doc['type'] = [f'x{len(self.hierStack)}']
      if not hierStack:
        hierStack = self.hierStack
      if 'id' not in doc:
        doc['id'] = hierStack[-1]
      if len(hierStack)>0 and doc['type'][0][0]=='x':
        hierStack  = hierStack[:-1]
        oldPath    =  doc['branch'][0]['path']
      elif 'branch' in doc:
        hierStack   = doc['branch'][0]['stack']
    else:  #new doc
      edit = False
      doc['type'] = docType.split('/')
      if len(hierStack) == 0:
        hierStack = self.hierStack
    logging.debug('Add/edit data in cwd:%s with stack:%s and name: %s and type: %s and edit: %s',self.cwd, hierStack, doc['name'], doc['type'], edit)
    # collect structure-doc and prepare
    if doc['type'][0] and doc['type'][0][0]=='x' and doc['type'][0]!='x0' and childNum is None:
      #should not have childnumber in other cases
      thisStack = '/'.join(hierStack)+'/'
      view = self.db.getView('viewHierarchy/viewHierarchy', startKey=thisStack) #not faster with cT.getChildren
      childNum = 0
      for item in view:
        if item['value'][1][0][0]!='x':
          continue
        if thisStack == '/'.join(item['key'].split('/')[:-1])+'/': #remove last item from string
          childNum += 1

    # find path name on local file system; name can be anything
    if self.cwd is not None and 'name' in doc:
      if doc['type'][0] and doc['type'][0][0]=='x':
        #project, step, task
        if doc['type'][0]=='x0':
          childNum = 0
        #parentDir
        #  edit: cwd of the project/step/task: remove last directory from cwd (since cwd contains a / at end: remove two)
        #  new: below the current project/step/task
        parentDirectory = self.cwd.parent if edit else self.cwd
        operation = 'u'
        # prevent that projects have the same name: #TODO make this a separate function and merge others
        dirNameBase = createDirName(doc['name'],doc['type'][0],childNum)
        dirName = dirNameBase
        if 'branch' not in doc or doc['branch'][0]['path'] != dirName: #only change if it is not the same name as before
          idx = 0
          while (parentDirectory/dirName).exists():
            idx += 1
            dirName = f'{dirNameBase}_{idx:02d}'
        path = parentDirectory/dirName #update,or create (if new doc, update ignored anyhow)
      else:
        #measurement, sample, procedure
        shasum = ''
        if '://' in doc['name']:                           #make up name
          if localCopy:
            baseName  = Path(doc['name']).stem
            extension = Path(doc['name']).suffix
            path = self.cwd/(camelCase(baseName)+extension)
            request.urlretrieve(doc['name'], path)
            doc['name'] = camelCase(baseName)+extension
          else:
            path = Path(doc['name'])
            try:
              shasum  = generic_hash(path)
            except Exception:
              print('**ERROR bad01: fetch remote content failed. Data not added')
              return {'id':''}
        elif doc['name']!='' and (self.basePath/doc['name']).is_file():          #file exists
          path = self.basePath/doc['name']
          doc['name'] = Path(doc['name']).name
        elif doc['name']!='' and (self.cwd/doc['name']).is_file():               #file exists
          path = self.cwd/doc['name']
        elif 'branch' in doc:
          if len(doc['branch'])==1:
            if doc['branch'][0]['path'] is not None and (self.basePath/doc['branch'][0]['path']).is_file():
              path = self.basePath/doc['branch'][0]['path']
          else:
            logging.warning('backend - known issue: add/edit document with multiple branches %s.', doc['id'])
            # I might change the wrong one if I change the branch, but there is nothing in table that can distinguish which branch to change
        else:                                                                     #make up name
          shasum  = '-'
        if shasum!='-' and path is not None:
          if shasum == '':
            shasum = generic_hash(path, forceFile=True)
          view = self.db.getView('viewIdentify/viewSHAsum',shasum)
          if len(view)==0 or forceNewImage:  #measurement not in database: create doc
            self.useExtractors(path,shasum,doc)  #create image/content
            # All files should appear in database
            # if not 'image' in doc and not 'content' in doc and not 'otherELNName' in doc:  #did not get valuable data: extractor does not exit
            #   return ''
          if len(view)==1:  #measurement is already in database
            doc['id'] = view[0]['id']
            doc['shasum'] = shasum
            edit = True
    # assemble branch information
    if childNum is None:
      childNum=9999
    if path is not None and path.is_absolute():
      path = path.relative_to(self.basePath)
    pathStr = None if path is None else path.as_posix().replace(':/','://') if path.as_posix().startswith('http') else path.as_posix()
    show = [True]*(len(hierStack)+1)
    if 'branch' in doc and len(hierStack)+1==len(doc['branch'][0]['show']):
      show = doc['branch'][0]['show']
    doc['branch'] = {'stack':hierStack,'child':childNum,'path':pathStr, 'show':show, 'op':operation}
    if edit:
      #update document
      keysNone = [key for key in doc if doc[key] is None]
      doc = fillDocBeforeCreate(doc, ['--'])
      for key in keysNone:
        doc[key]=None
      doc = self.db.updateDoc(doc, doc['id'])
    else:
      # add doc to database
      doc = fillDocBeforeCreate(doc, doc['type'])
      doc = self.db.saveDoc(doc)
    ## adaptation of directory tree, information on disk: documentID is required
    if self.cwd is not None and doc['type'][0][0]=='x':
      #project, step, task
      path = Path(doc['branch'][0]['path'])
      if edit and oldPath is not None:
        if not (self.basePath/oldPath).is_dir():
          print(f'**WARNING: addData edit of folder should have oldPath and that should exist:{oldPath}'
                f'\n This can be triggered if user moved the folder.')
          return  {'id':''}
        (self.basePath/oldPath).rename(self.basePath/path)
      else:
        (self.basePath/path).mkdir(exist_ok=True)   #if exist, create again; moving not necessary since directory moved in changeHierarchy
      with open(self.basePath/path/'.id_pastaELN.json','w', encoding='utf-8') as f:  #local path, update in any case
        f.write(json.dumps(doc))
    return doc


  ######################################################
  ### Disk directory/folder methods
  ######################################################
  def changeHierarchy(self, docID:Optional[str], dirName:Optional[Path]=None) -> None:
    """
    Change through text hierarchy structure
    change hierarchyStack, change directory, change stored cwd

    Args:
        docID (string): information on how to change
        dirName (string): change into this directory (absolute path given). For if data is moved
    """
    logging.debug('changeHierarchy should only be used in CLI mode')
    if self.cwd is None:
      return
    if docID is None or (docID[0]=='x' and docID[1]!='-'):  #cd ..: none. close 'project', 'task'
      self.hierStack.pop()
      self.cwd = self.cwd.parent
    elif dirName is None:  # existing ID is given: open that
      doc = self.db.getDoc(docID)
      self.cwd = self.basePath/doc['branch'][0]['path']
      self.hierStack = doc['branch'][0]['stack']+[docID]
    else:
      self.cwd = dirName
      self.hierStack.append(docID)
    return


  def scanProject(self, progressBar:Callable[...,None]|None, projID:str, projPath:Path|None=None) -> None:
    """ Scan directory tree recursively from project/... or project/task/....
    - find changes on file system and move those changes to DB
    - use .id_pastaELN.json to track changes of directories, aka projects/steps/tasks
    - use shasum to track changes of measurements etc. (one file=one shasum=one entry in DB)
    - create database entries for measurements in directory
    - move/copy/delete allowed as the doc['path'] = list of all copies
      doc['path'] is adopted once changes are observed

    Args:
      progressBar (func): progress bar
      projID (str): project's docID
      projPath (str): project's path from basePath; if not given, will be determined

    Raises:
      ValueError: could not add new measurement to database
    """
    rerunScanTree = False
    self.hierStack = [projID]
    if projPath is None:
      pathPosix:str = self.db.getDoc(projID)['branch'][0]['path']
      self.cwd      = self.basePath/pathPosix
      projPath      = self.cwd.relative_to(self.basePath)
    else:
      self.cwd = self.basePath/projPath
    #prepare lists and start iterating
    inDB_all = self.db.getView('viewHierarchy/viewPathsAll', startKey=projPath.as_posix())
    pathsInDB_x    = [i['key'] for i in inDB_all if i['value'][1][0][0]=='x']  #all structure elements: task, subtasts
    pathsInDB_data = [i['key'] for i in inDB_all if i['value'][1][0][0]!='x']
    filesCountSum = sum(len(files) for (_, _, files) in os.walk(self.cwd))
    filesCount = 0
    for root, dirs, files in os.walk(self.cwd, topdown=True):
      #find parent-document
      self.cwd = Path(root).relative_to(self.basePath)
      if self.cwd.name.startswith('trash_'):
        del dirs
        del files
        continue
      parentIDs = [i for i in inDB_all if i['key']==self.cwd.as_posix()]  #parent of this folder
      if not parentIDs: #skip newly moved folder, will be scanned upon re-scanning
        continue
      parentID = parentIDs[0]['id']
      parentDoc = self.db.getDoc(parentID)
      hierStack = parentDoc['branch'][0]['stack']+[parentID]
      # handle directories
      dirs[:] = [i for i in dirs if not i.startswith(('.','trash_')) and i not in ('__pycache__')
                 and not (Path(root)/i/'pyvenv.cfg').is_file()]
      for dirName in dirs[::-1]: #sorted forward in Linux
        path = (Path(root)/dirName).relative_to(self.basePath).as_posix()
        if path in pathsInDB_x: #path already in database
          pathsInDB_x.remove(path)
          continue
        if (self.basePath/path/'.id_pastaELN.json').is_file(): # update branch: path and stack
          with open(self.basePath/path/'.id_pastaELN.json', encoding='utf-8') as fIn:
            doc = json.loads(fIn.read())
            doc = self.db.getDoc(doc['id'])
          if len(doc)==0:
            (self.basePath/path/'.id_pastaELN.json').unlink()
            rerunScanTree = True
            continue
          if (self.basePath/doc['branch'][0]['path']).parent.as_posix()  == root and \
               doc['branch'][0]['stack']==hierStack:
            # special case: user wants to have a different directory name in same folder: then the child-number should not change
            childNum = doc['branch'][0]['child']
            newPath = path
          else:
            #determine childNumber
            thisStack = '/'.join(hierStack)
            view = self.db.getView('viewHierarchy/viewHierarchy', startKey=thisStack)
            childNum = 0
            for item in view:
              if item['value'][1][0]=='x0' or item['value'][1][0][0]!='x':
                continue
              if thisStack == ' '.join(item['key'].split(' ')[:-1]): #remove last item from string
                childNum += 1
            newPath = '/'.join(path.split('/')[:-1])+'/'+createDirName(doc['name'],doc['type'][0],childNum) #update,or create (if new doc, update ignored anyhow)
            if (self.basePath/newPath).exists():                     #can be either file or directory
              print('**ERROR new path should not exist',newPath)
            else:
              (self.basePath/path).rename(self.basePath/newPath)
          self.db.updateBranch(doc['id'], 0, childNum, hierStack, newPath)
        else:
          currentID = self.addData('x1', {'name': dirName}, hierStack)['id']
          newDir = self.basePath/self.db.getDoc(currentID)['branch'][0]['path']
          (newDir/'.id_pastaELN.json').rename(self.basePath/root/dirName/'.id_pastaELN.json') #move index file into old folder
          newDir.rmdir()                     #remove created path
          (self.basePath/root/dirName).rename(newDir) #move old to new path
        rerunScanTree = True
      # handle files
      for fileName in files:
        filesCount += 1
        if progressBar is not None:
          progressBar(int(100*filesCount/filesCountSum))
        if fileName.startswith(('.', 'trash_')) or '_PastaExport' in fileName:  #ignore files
          continue
        path = (Path(root).relative_to(self.basePath) /fileName).as_posix()
        if path in pathsInDB_data:
          logging.info('Scan: file already in DB: %s',path)
          pathsInDB_data.remove(path)
        else:
          logging.info('Scan: add file to DB: %s',path)
          shasum = generic_hash(self.basePath/path, forceFile=True)
          if not shasum:
            raise NameError(f'Filepath does not exist {self.basePath/path}')
          view = self.db.getView('viewIdentify/viewSHAsum',shasum)
          if len(view)==0:                             #not in database: create doc
            self.addData('', {'name':path}, hierStack)
          else:
            self.db.updateBranch(view[0]['id'], -1, 9999, hierStack, path)
    #finish method
    self.cwd = self.basePath/projPath
    orphans = [
        i for i in pathsInDB_data
        if i.startswith(f'{self.cwd.relative_to(self.basePath).as_posix()}/')
    ]
    logging.info('Scan: these files are on DB but not hard disk\n%s','\n  '.join(orphans))
    orphanDirs = [
        i for i in pathsInDB_x
        if i.startswith(f'{self.cwd.relative_to(self.basePath).as_posix()}/')
        and i != projPath
    ]
    logging.info('Scan: these directories are on DB but not hard disk\n%s','\n  '.join(orphanDirs))
    for orphan in orphans+orphanDirs:
      docID = [i for i in inDB_all if i['key']==orphan][0]['id']
      self.db.updateBranch(docID, -2, 9999, [], orphan)
    #reset to initial values
    self.hierStack = []
    self.cwd = Path(self.basePath)
    if rerunScanTree:
      self.scanProject(progressBar, projID, projPath)
    return


  def useExtractors(self, filePath:Path, shasum:str, doc:dict[str,Any]) -> None:
    """
    get measurements from datafile: central distribution point
    - max image size defined here

    Args:
        filePath (Path): path to file
        shasum (string): shasum (git-style hash) to store in database (not used here)
        doc (dict): pass known data/measurement type, can be used to create image; This doc is altered
    """
    extension = filePath.suffix[1:]  #cut off initial . of .jpg
    if str(filePath).startswith('http'):
      absFilePath = Path(tempfile.gettempdir())/filePath.name
      with request.urlopen(filePath.as_posix().replace(':/','://'), timeout=60) as urlRequest:
        with open(absFilePath, 'wb') as f:
          try:
            f.write(urlRequest.read())
          except Exception:
            print('Error saving downloaded file to temporary disk')
    else:
      if filePath.is_absolute():
        filePath = filePath.relative_to(self.basePath)
      absFilePath = self.basePath/filePath
    pyFile = f'extractor_{extension.lower()}.py'
    pyPath = self.addOnPath/pyFile
    if pyPath.is_file():
      # import module and use to get data
      os.environ['QT_API'] = 'pyside2'
      import matplotlib.pyplot as plt  # IMPORTANT: NO PYPLOT OUTSIDE THIS QT_API BLOCK
      plt.clf()
      try:
        module  = importlib.import_module(pyFile[:-3])
        content = module.use(absFilePath, {'main':'/'.join(doc['type'])} )
        for key in [i for i in content if i not in ['metaVendor','metaUser','image','content','links','style']]:  #only allow accepted keys
          del content[key]
        doc |= content
        for meta in ['metaVendor','metaUser']:
          if meta not in doc:
            doc[meta] = {}
          if isinstance(doc[meta], dict):  #convenient type
            for item in doc[meta]:
              if isinstance(doc[meta][item], tuple):
                doc[meta][item] = list(doc[meta][item])
              try:
                _ = json.dumps(doc[meta][item])
              except (ValueError, TypeError):
                doc[meta][item] = str(doc[meta][item])
                print('**Warning -> stringified  ',meta, item)
          else:
            for item in doc[meta]:
              if not (isinstance(item, dict) and 'key' in item and 'value' in item and 'unit' in item):
                logging.error('Complicated extractor return wrong')
                print('Complicated extractor return wrong')
        if doc['style']['main'].startswith(doc['type'][0]):
          doc['type']     = doc['style']['main'].split('/')
        else:
          #user has strange wish: trust him/her
          logging.info('user has strange wish: trust him/her: %s  %s','/'.join(doc['type']),'  '+doc['style']['main'])
        del doc['style']
        if 'fileExtension' not in doc['metaVendor']:
          doc['metaVendor']['fileExtension'] = extension.lower()
        if 'links' in doc and len(doc['links'])==0:
          del doc['links']
      except Exception:
        print(f'  **Warning, issue with extractor {pyFile} {absFilePath}')
        logging.warning('Issue with extractor %s\n %s', pyFile, traceback.format_exc())
        doc['metaUser'] = {'filename':absFilePath.name, 'extension':absFilePath.suffix,
          'filesize':absFilePath.stat().st_size,
          'created at':datetime.fromtimestamp(absFilePath.stat().st_ctime, tz=timezone.utc).isoformat(),
          'modified at':datetime.fromtimestamp(absFilePath.stat().st_mtime, tz=timezone.utc).isoformat()}
    #combine into document
    os.environ['QT_API'] = 'pyside6'
    doc['shasum']=shasum  #essential for logic, always save, unlike image
    return


  def testExtractor(self, filePath:Union[Path,str], extractorPath:Optional[Path]=None, style:dict[str,Any]={'main':''},
                    outputStyle:str='text', saveFig:str='') -> str:
    """
    Args:
      filePath (Path, str): path to the file to be tested
      extractorPath (Path, None): path to the directory with extractors
      style (dict): style with a main-key that is / separated
      outputStyle (str): report in ['print','text','html'] including show images
      saveFig (str): save figure to...; if given stop testing after generating image

    Returns:
      str: short summary or long report
    """
    import base64
    from io import BytesIO
    import cairosvg
    from PIL import Image
    os.environ['QT_API'] = 'pyside2'
    import matplotlib.axes as mpaxes
    import matplotlib.pyplot as plt

    htmlStr= 'Please visit <a href="https://pasta-eln.github.io/pasta-eln/extractors.html#'
    success = True
    if isinstance(filePath, str):
      filePath = Path(filePath)
    if filePath.as_posix().startswith('http'):
      tempFilePath = Path(tempfile.gettempdir())/filePath.name
      request.urlretrieve(filePath.as_posix().replace(':/','://'), tempFilePath)
      filePath = tempFilePath
    report = outputString(outputStyle, 'h2', 'Report on extractor test')
    report += outputString(outputStyle, 'info', f'check file: {str(filePath)}')
    extension = filePath.suffix[1:]
    pyFile = f'extractor_{extension.lower()}.py'
    if extractorPath is None:
      extractorPath = self.addOnPath
    #start testing
    if (extractorPath/pyFile).is_file():
      report += outputString(outputStyle, 'info', f'use extractor: {str(extractorPath / pyFile)}')
    else:
      success = False
      report += outputString(outputStyle, 'error', f'No fitting extractor found:{pyFile}')
    if success:
      try:
        module  = importlib.import_module(pyFile[:-3])
        plt.clf()
        content = module.use(filePath, style, saveFig or None )
        if saveFig:
          return report
      except Exception:
        success = False
        report += outputString(outputStyle, 'error', 'Python error in extractor')
        report += outputString(outputStyle, 'error', f'{htmlStr}python-error">website</a>')
        report += outputString(outputStyle, 'error', traceback.format_exc(limit=3))
    if success:
      if 'style' in content:
        possibleDocTypes = [i for i in self.db.dataHierarchy('', '') if i[0]!='x']
        matches = [i for i in possibleDocTypes if content['style']['main'].startswith(i)]
        if matches or content['style']['main'] in {'', '-'}:
          report += outputString(outputStyle, 'info', 'Style is good: '+content['style']['main'])
          size = len(str(content))
          report += outputString(outputStyle, 'info', f'Entire extracted size: {size // 1024}kB')
        else:
          report += outputString(outputStyle, 'error', 'Style does not follow doctype in dataHierarchy.')
      else:
        report += outputString(outputStyle,'error','Style not included in extractor.')
    if success:
      try:
        _ = json.dumps(content)
      except Exception:
        report += outputString(outputStyle,'error','Extractor reply not json dumpable.')
    if success:
      try:
        _ = json.dumps(content['metaVendor'])
        if not isinstance(content['metaVendor'], (dict,list)):
          raise TypeError(' Meta vendor: wrong type')
        report += outputString(outputStyle,'info','Number of vendor entries: '+str(len(content['metaVendor'])))
      except Exception:
        # possible cause of failure: make sure that no int64 but normal int
        success = False
        report += outputString(outputStyle,'error', 'Some json format does not fit in metaVendor')
        report += outputString(outputStyle, 'error', f'{htmlStr}metadata-error">website</a>')
        #iterate keys
        for key in content['metaVendor']:
          try:
            _ = json.dumps(content['metaVendor'][key])
          except Exception:
            report += outputString(outputStyle,'error',f'FAIL {key}:'+str(content['metaVendor'][key])+' type:')+str(type(content['metaVendor'][key]))
    if success:
      try:
        _ = json.dumps(content['metaUser'])
        if not isinstance(content['metaUser'], (dict,list)):
          raise TypeError(' Meta user: wrong type')
        report += 'Number of user entries: '+str(len(content['metaUser']))+'<br>'
      except Exception:
        report += outputString(outputStyle,'error', 'Some json format does not fit in metaUser')
        report += outputString(outputStyle, 'error', f'{htmlStr}metadata-error">website</a>')
        #iterate keys
        for key in content['metaUser']:
          try:
            _ = json.dumps(content['metaUser'][key])
          except Exception:
            report += outputString(outputStyle,'error',f'FAIL {key}:'+str(content['metaUser'][key])+' type:') + str(type(content['metaUser'][key]))
      if 'image' not in content:
        success = False
        report += outputString(outputStyle,'error','Image does not exist')
    if success and isinstance(content['image'],Image.Image):
      success = False
      report += outputString(outputStyle,'error','Image is a PIL image: not a base64 string')
      report += outputString(outputStyle, 'error', f'{htmlStr}pillow-image">website</a>')
        # print('Encode image via the following: pay attention to jpg/png which is encoded twice\n```')
        # print('from io import BytesIO')
        # print('figfile = BytesIO()')
        # print('image.save(figfile, format="PNG")')
        # print('imageData = base64.b64encode(figfile.getvalue()).decode()')
        # print('image = "data:image/jpg;base64," + imageData')
    if success and isinstance(content['image'], mpaxes._axes.Axes): # pylint: disable=protected-access
      success = False
      report += outputString(outputStyle,'error','Image is a matplotlib axis: not a base64 string')
      report += outputString(outputStyle, 'error', f'{htmlStr}matplotlib">website</a>')
        # print('**Warning: image is a matplotlib axis: not a svg string')
        # print('  figfile = StringIO()')
        # print('plt.savefig(figfile, format="svg")')
        # print('image = figfile.getvalue()')
    if success and isinstance(content['image'], str):#show content
      size = len(content['image'])
      report += outputString(outputStyle, 'info', f'Image size {str(size // 1024)}kB')
      if outputStyle!='text':
        report += outputString(outputStyle,'h2','Additional window shows the image')
      if len(content['image'])>20:
        if content['image'].startswith('data:image/'):
          #png or jpg encoded base64
          extension = content['image'][11:14]
          img = base64.b64decode(content['image'][22:])
        else:
          #svg data
          img = cairosvg.svg2png(bytestring=content['image'].encode())
        i = BytesIO(img)
        image = Image.open(i)
        if outputStyle!='text':
          image.show()
      del content['image']
    if outputStyle=='print':
      print('Identified metadata',content)
    os.environ['QT_API'] = 'pyside6'
    return report


  ######################################################
  ### Wrapper for database functions
  ######################################################
  def checkDB(self, outputStyle:str='text', repair:Union[None,Callable[[str],bool]]=None,
              minimal:bool=False) -> str:
    """
    Wrapper of check database for consistencies by iterating through all documents

    Args:
        outputStyle (str): output using a given style: see outputString
        repair (function): repair errors automatically; function that has user interaction
        minimal (bool): true=only show warnings and errors; else=also show information

    Returns:
        string: output incl. \n
    """
    ### check database itself for consistency
    output = self.db.checkDB(outputStyle=outputStyle, minimal=minimal, repair=repair)
    ### compare with file system
    if not minimal:
      output += outputString(outputStyle,'h2','File status')
    viewProjects   = self.db.getView('viewDocType/x0All')
    inDB_all = self.db.getView('viewHierarchy/viewPathsAll')
    pathsInDB_data = [i['key'] for i in inDB_all if i['value'][1][0][0]!='x']
    pathsInDB_folder = [i['key'] for i in inDB_all if i['value'][1][0][0]=='x']
    count = 0
    for projI in viewProjects['id']:
      projDoc = self.db.getDoc(projI)
      if len(projDoc['branch'])==0:
        output += outputString(outputStyle,'error','project view got screwed up')
        continue
      for root, dirs, files in os.walk(self.basePath/projDoc['branch'][0]['path']):
        if Path(root).name[0]=='.' or Path(root).name.startswith('trash_'):
          del dirs
          continue
        for fileName in files:
          if fileName.startswith('.') or fileName.startswith('trash_') or '_PastaExport' in fileName:
            continue
          path = (Path(root).relative_to(self.basePath) /fileName).as_posix()
          if path not in pathsInDB_data:
            output += outputString(outputStyle, 'error', f'File   on disk but not DB (2): {path}')
            count += 1
          else:
            pathsInDB_data.remove(path)
        dirs[:] = [i for i in dirs if not i.startswith(('.','trash_')) and i not in ('__pycache__')
                  and not (Path(root)/i/'pyvenv.cfg').is_file()]
        for dirName in dirs:
          path = (Path(root).relative_to(self.basePath) /dirName).as_posix()
          if path not in pathsInDB_folder:
            output += outputString(outputStyle, 'error', f'Folder on disk but not DB    : {path}')
            count += 1
          else:
            pathsInDB_folder.remove(path)
            listDocs = self.db.getView('viewHierarchy/viewPathsAll', preciseKey=path)
            if len(listDocs)!=1:
              output += outputString(outputStyle, 'error', f'Path of folder is non-unique (1): {path} in '\
                                      f'{" ".join([i["id"] for i in listDocs])}')
            docDB   = self.db.getDoc(listDocs[0]['id'])
            if (self.basePath/root/dirName/'.id_pastaELN.json').is_file():
              with open(self.basePath/root/dirName/'.id_pastaELN.json',encoding='utf-8') as fIn:
                docDisk = json.loads(fIn.read())
                difference = diffDicts(docDisk,docDB)
                if len(difference)>1:
                  errorStr = outputString(outputStyle,'error',f'disk(1) and db(2) content do not match*: {docDB["id"]}\n{difference}')
                  if repair is None:
                    output += errorStr
                  elif repair(errorStr):
                    with open(self.basePath/root/dirName/'.id_pastaELN.json','w',encoding='utf-8') as fOut:
                      json.dump(docDB, fOut)
                  # #use only for resetting the content in the .id_pastaELN.json
            else:
              count += 1
              errorStr = outputString(outputStyle, 'error', f'Folder has no .id_pastaELN.json:{path}')
              if repair is None:
                output += errorStr
              elif repair(errorStr):
                with open(self.basePath/root/dirName/'.id_pastaELN.json','w',encoding='utf-8') as fOut:
                  json.dump({'id':docDB['id']}, fOut)
    orphans = [i for i in pathsInDB_data   if not (self.basePath/i).exists() and ':/' not in i and i!='*']  #paths can be files or directories
    orphans+= [i for i in pathsInDB_folder if not (self.basePath/i).exists() ]
    if orphans:
      if repair is None:
        output += outputString(outputStyle,'error','bch01: These paths of database not on filesystem(3):\n  - '+'\n  - '.join(orphans))
      else:
        for orphan in sorted(orphans):
          self.db.cursor.execute(f'SELECT main.name, main.type, branches.path, main.id, main.comment FROM main JOIN branches USING(id) WHERE branches.path == "{orphan}"')
          res = self.db.cursor.fetchall()
          resString = '\n  '.join(str(i) for i in res)
          if repair(f'Path of database not on filesystem:\n  {resString}. Repair: file-remove path; folder-create folder and .id_pastaELN'):
            if res[0][1].startswith('x'):
              (self.basePath/orphan).mkdir(parents=True)
              with open(self.basePath/orphan/'.id_pastaELN.json','w',encoding='utf-8') as fOut:
                json.dump({'id':res[0][3]}, fOut)
            else:
              self.db.cursor.execute(f"UPDATE branches SET path='*' WHERE id == '{res[0][3]}' and path == '{orphan}'")
              self.db.connection.commit()
    if not minimal:
      output += outputString(outputStyle,'h2','File summary')
    if outputStyle == 'text':
      output += 'Success\n' if not orphans and count==0 else 'Failure (* can be auto-repaired)\n'
    return output



"""extract data from vendor
- Log  .log for procedures
For tutorials, see extractor_csv.py and extractor_png.py
"""

def use(filePath, style={'main':''}, saveFileName=None):
  """
  Args:
    filePath (Path): full path file name
    style    (dict): supplied to guide the display / extraction style = recipe
                     main is / separated hierarchical elements parent->child
                     can contain more elements
    saveFileName (string): if given, save the image to this file-name
  Returns:
    dict: containing image, metaVendor, metaUser, recipe
  """
  with open(filePath, encoding='utf-8') as fIn:
    return {'content':fIn.read(), 'style': {'main':'workflow/worklog'}, 'metaVendor':{}, 'metaUser':{}}



""" Allow syncing to elabFTW server """
import copy
import json
import logging
import re
import traceback
from collections import Counter
from datetime import datetime
from typing import Any, Callable
from anytree import Node, PreOrderIter
from .backend import Backend
from .elabFTWapi import ElabFTWApi
from .miscTools import flatten
from .textTools.handleDictionaries import squashTupleIntoValue
from .textTools.html2markdown import html2markdown
from .textTools.markdown2html import markdown2html  # type: ignore[attr-defined]

# - consider hiding metadata.json (requires hiding the upload (state=2) and ability to read (it is even hidden in the API-read))
#   - hide an upload  api.upLoadUpdate('experiments', 66, 596, {'action':'update', 'state':'2'})
#   - listing all uploads (incl. archived) is not possible in elab currently -> leave visible; change to invisible once in elab

def cliCallback(api:ElabFTWApi , entry:str, idx:int) -> str:
  """ Default callback function for the syncMissingEntries function

  Args:
    api (ElabFTWApi): api to ask about more information
    entry (str): entry type to ask about
    idx (int): entry number on elabFTW server

  Returns:
    str: mode of processing: g,gA,s,sA
  """
  print(f'**ERROR**: default callback function should not be called. API:{api}, entry:{entry}, idx:{idx}')
  return ''


MERGE_LABELS = {
     -1:'-1: ERROR',
      1:'1: client -> server',
      2:'2: other client -> client',
      3:'3: server -> client',
      4:'4: no change',
      5:'5: BOTH CHANGE / Merge'
    }


class Pasta2Elab:
  """ Allow syncing to elabFTW server"""

  def __init__(self, backend:Backend, projectGroup:str='', purge:bool=False):
    '''
    initiate an elab instance to allow for syncing: save data-hierarchy information
    - cannot save pasta's measurement data hierarchy in any other docTypes; hence, makes no sense to save any data hierarchy
      into any elab-item-types metadata

    Args:
      backend (backend): backend
      projectGroup (str): name of project group
      purge (bool): remove old documents

    Returns:
      bool: success
    '''
    self.backend = backend
    configuration = self.backend.configuration
    self.projectGroup = projectGroup or configuration['defaultProjectGroup']
    if not configuration['projectGroups'][self.projectGroup]['remote']['url'] or \
       not configuration['projectGroups'][self.projectGroup]['remote']['key'] or \
       'config' not in configuration['projectGroups'][self.projectGroup]['remote'] or \
       'id' not in configuration['projectGroups'][self.projectGroup]['remote']['config']:
      return
    self.api = ElabFTWApi(configuration['projectGroups'][self.projectGroup]['remote']['url'],
                          configuration['projectGroups'][self.projectGroup]['remote']['key'])
    configRemote = configuration['projectGroups'][self.projectGroup]['remote']
    self.elabProjGroupID = configRemote['config']['id']
    if purge:
      data = self.api.readEntry('items', self.elabProjGroupID)[0]
      _ = [self.api.deleteEntry('experiments', i['entityid']) for i in data['related_experiments_links']]
      _ = [self.api.deleteEntry('items',       i['entityid']) for i in data['related_items_links']]
    self.docID2elabID:dict[str,tuple[int,bool]] = {}  # x-15343154th54325243, (4, bool if experiment)
    self.readWriteAccess:dict[str,str] = {}
    self.verbose         = False
    return


  def sync(self, mode:str='', callback:Callable[[ElabFTWApi,str,int],str]=cliCallback,
           progressCallback:Callable[...,None]|None=None) -> list[tuple[str,int]]:
    """ Main function

    Args:
      mode (str): sync mode g=get, gA=get-all, s=send, sA=send-all
      callback (func): callback function if non-all mode is given
      progressCallback (func): callback function to implement progress-bar

    Returns:
      list: list of merge cases
    """
    def updateEntryLocal(i:Node, mode:str, callback:Callable[[ElabFTWApi,str,int],str]=cliCallback,
                         idx:int=-1, count:int=-1) -> tuple[str,int]:
      """Intermediate function used in list comprehension"""
      res = self.updateEntry(i, mode, callback)
      if progressCallback is not None:
        progressCallback('count', str(int(idx/count*100)))
      return res

    if hasattr(self,'api') and self.api.url:  #only when you are connected to web
      report = []
      if progressCallback is not None:
        progressCallback('text', '### Start syncing with elabFTW server\n#### Set up sync\nStart...')
      self.syncDocTypes()  # sync categories ~1sec
      self.createIdDict()  # TODO: get progressCallback as argument
      if progressCallback is not None:
        progressCallback('append', 'Done\n#### Sync each document\nStart...')
      for projID in self.backend.db.getView('viewDocType/x0')['id'].values:
        projHierarchy, _ = self.backend.db.getHierarchy(projID)
        count = len(list(PreOrderIter(projHierarchy)))
        report += [updateEntryLocal(i, mode, callback, idx, count) for idx, i in enumerate(PreOrderIter(projHierarchy))]
      if progressCallback is not None:
        progressCallback('append', 'Done\n#### Sync missing entries\nStart...')
      report += self.syncMissingEntries(mode, callback, progressCallback)
    else:
      print('**ERROR Not connected to elab server!')
      return []
    if progressCallback is not None:
      reportSum = Counter([i[1] for i in report])
      reportText = '\n  - '.join(['']+[f'{v:>4}:{MERGE_LABELS[k][2:]}' for k,v in reportSum.items()])
      progressCallback('count', '100')
      progressCallback('append', f'Done\n#### Summary\nSend all data to server: success\n{reportText}')
    return report


  def syncDocTypes(self) -> None:
    """ Synchronize document types between client and server
    - save datahierarchy to server
    """
    #try body and metadata: compare, write
    docTypesElab  = {i['title']:i['id'] for i in self.api.readEntry('items_types')}
    docTypesPasta = {i.capitalize() for i in self.backend.db.dataHierarchy('','') if not i.startswith('x')} | \
                    {'Default','Folder','Project','ProjectGroup'}
    for docType in docTypesPasta.difference({'Measurement'}|docTypesElab.keys()):  # do not create measurements, use 'experiments'
      self.api.touchEntry('items_types', {'title': docType})
    #verify nothing extraneous
    docTypesElab  = {i['title']:i['id'] for i in self.api.readEntry('items_types')}
    if set(docTypesElab.keys()).difference(docTypesPasta|{'Default','ProjectGroup'}) and self.verbose:
      print('**Info: some items exist that should not:', set(docTypesElab.keys()).difference(docTypesPasta|{'Default'}),
            'You can remove manually, but should not interfere since not used.')
    # sync data-hierarchy
    dataHierarchy = []
    for docType in self.backend.db.dataHierarchy('',''):
      dataHierarchy += copy.deepcopy([dict(i) for i in self.backend.db.dataHierarchy(docType,'meta')])
    self.api.updateEntry('items', self.elabProjGroupID, {'metadata':json.dumps(dataHierarchy)})
    return


  def createIdDict(self) -> None:
    """ create mapping of docIDs to elabIDs: if not exist, create elabIds """
    configRemote = self.backend.configuration['projectGroups'][self.projectGroup]['remote']['config']
    # set default read/write access rules
    self.readWriteAccess =  {'canwrite':configRemote['canWrite'], 'canread':configRemote['canRead']}
    elabTypes = {i['title'].lower():i['id'] for i in self.api.readEntry('items_types')}  |  {'measurement':-1}
    elabTypes |= {'x0':elabTypes.pop('project'), 'x1':elabTypes.pop('folder'), '-':elabTypes.pop('default')}
    def getNewEntry(elabType:str) -> int:
      urlSuffix = 'items'                  if int(elabType)>0 else 'experiments'
      content   = {'category_id':elabType} | self.readWriteAccess
      elabID    = self.api.touchEntry(urlSuffix, content)
      self.api.createLink(urlSuffix, elabID, 'items', self.elabProjGroupID)
      return elabID
    self.backend.db.cursor.execute('SELECT id, type, externalId FROM main')
    self.docID2elabID = {i[0]:((i[2],i[1].split('/')[0]=='measurement') if i[2] else (getNewEntry(elabTypes[i[1].split('/')[0]]),i[1].split('/')[0]=='measurement'))
                    for i in self.backend.db.cursor.fetchall()}
    if self.verbose:
      print('List of docIDs and corresponding elabIDs (flag if experiment)')
      print('\n'.join([f'{k} : {v}' for k,v in self.docID2elabID.items()]))
    # save to sqlite
    self.backend.db.cursor.executemany('UPDATE main SET id=?, externalId=? WHERE id=?', [(k,v[0],k) for k, v in self.docID2elabID.items()])
    self.backend.db.connection.commit()
    return


  def updateEntry(self, node:Node, mode:str, callback:Callable[[ElabFTWApi,str,int],str]=cliCallback) -> tuple[str,int]:
    """ update an entry in elabFTW: all the logic goes here
        - myDesktop: sends content and the date when upload is made; if there is a change in modified time; the change is for real
        - server: gets document; if there is a difference between metadata.json and content: it was changed for real
        - other desktops changes: as before; my desktop **cannot differentiate myChanges vs other desktop changes**
          -> have to save also upload time locally, save in extra column

    Args:
      node (Node): node to process
      mode (str): sync mode g=get, gA=get-all, s=send, sA=send-all
      callback (func): callback function if non-all mode is given

    Returns:
      tuple: node.id; merge case
    """
    # get this content: check if it changed
    docClient   = self.backend.db.getDoc(node.id)
    elabID = self.docID2elabID[node.id][0]
    if 'dateSync' not in docClient or not docClient['dateSync']:
      docClient['dateSync'] = datetime.fromisoformat('2000-01-03').isoformat()+'.0000'
    if self.verbose:
      print(f'\n{node.id}\n>>>DOC_CLIENT sync&modified', docClient['dateSync'], docClient['dateModified'])
    # pull from server: content and other's client content
    entryType = 'experiments' if self.docID2elabID[node.id][1] else 'items'
    docServer, uploads = self.elab2doc(self.api.readEntry(entryType, elabID)[0])
    if self.verbose:
      print('>>>DOC_SERVER', docServer)
    if listDoNotChange :=[i for i in uploads if i['real_name']=='do_not_change.json']:
      docOther = self.api.download(entryType, elabID, listDoNotChange[0])
    else:
      docOther = {'name':'Untitled', 'tags':[], 'comment':'', 'dateSync':datetime.fromisoformat('2000-01-02').isoformat()+'.0000',
                  'dateModified':datetime.fromisoformat('2000-01-01').isoformat()+'.0000'}
      # return(node.id, -1)
    if self.verbose:
      print('>>>DOC_OTHER sync&modified', docOther.get('dateSync'), docOther.get('dateModified'))
    docMerged:dict[str,Any] = {}
    flagUpdateClient, flagUpdateServer = False, False

    # merge 1: compare server content and docOther and update later with changes
    flagServerChange = False
    for k,v in docServer.items():
      if isinstance(v, str):
        if v != html2markdown(markdown2html(docOther[k])):
          flagServerChange = True
          if self.verbose:
            print(f'str change k:{k}; v:\n|{v}|\nvOther:\n|{docOther[k]}|\n|{html2markdown(markdown2html(docOther[k]))}|\ntype:{type(v)}')
          docOther[k] = docServer[k]
      elif isinstance(v, dict):
        if v != docOther[k]:
          flagServerChange = True
          if self.verbose:
            print(f'dict change k:{k}; v:{v}; vOther:{docOther[k]}|type:{type(v)}')
          docOther[k] = docServer[k]
      elif isinstance(v, list):
        if set(v) != {i for i in docOther[k] if not re.match(r'^_\d$', i)}:
          flagServerChange = True
          if self.verbose:
            print('list change', k,v, docOther[k])
          docOther[k] = docServer[k]
      elif self.verbose:
        print('other change', k,v, docOther[k], type(v))
    if flagServerChange:
      if self.verbose:
        print('Server content changed from other')
      docOther['dateModified'] = datetime.now().isoformat()

    # merge 2: compare server client
    pattern = '%Y-%m-%dT%H:%M:%S.%f'
    mergeCase = -1
    #  - Case 1 straight update from client to server: only client updated and server not changed
    if (datetime.strptime(docClient['dateModified'], pattern) >  datetime.strptime(docClient['dateSync'], pattern) and \
        datetime.strptime(docOther['dateModified'], pattern)  <= datetime.strptime(docOther['dateSync'], pattern) and  \
        not mode.startswith('g'))     or     mode=='sA':
      mergeCase = 1
      if self.verbose:
        print(f'** MERGE CASE {MERGE_LABELS[mergeCase]}')
      flagUpdateServer = True
      flagUpdateClient = False
      docMerged = copy.deepcopy(docClient)
    #  - Case 2 straight update from server to client: only others was updated / server itself was not updated, client not changed
    if (datetime.strptime(docClient['dateModified'], pattern) <= datetime.strptime(docClient['dateSync'], pattern) and \
        datetime.strptime(docOther['dateModified'], pattern)  >  datetime.strptime(docOther['dateSync'], pattern) and \
        not mode.startswith('s'))     or     mode=='gA':
      mergeCase = 2
      if self.verbose:
        print(f'** MERGE CASE {MERGE_LABELS[mergeCase]}')
      flagUpdateServer = False
      flagUpdateClient = True
      docMerged = copy.deepcopy(docOther)
    #  - Case 3 straight update from server to client: only server updated, client not changed
    if datetime.strptime(docClient['dateModified'], pattern) <= datetime.strptime(docClient['dateSync'], pattern) and \
       datetime.strptime(docOther['dateModified'], pattern)  >  datetime.strptime(docOther['dateSync'], pattern) and \
       flagServerChange and not mode.startswith('s'):
      mergeCase = 3
      if self.verbose:
        print(f'** MERGE CASE {MERGE_LABELS[mergeCase]}')
      flagUpdateServer = True
      flagUpdateClient = True
      docMerged = copy.deepcopy(docOther)
    #  - Case 4 no change: nothing changed
    if datetime.strptime(docClient['dateModified'], pattern) <= datetime.strptime(docClient['dateSync'], pattern) and \
       datetime.strptime(docOther['dateModified'], pattern)  <= datetime.strptime(docOther['dateSync'], pattern) and \
          not mode.endswith('A'):
      mergeCase = 4
      if self.verbose:
        print(f'** MERGE CASE {MERGE_LABELS[mergeCase]}')
      # flagUpdateServer = False
      # flagUpdateClient = False
      # docMerged = {}
      return node.id, mergeCase
    #  - Case 5 both are updated: merge: both changed -> GUI
    if datetime.strptime(docClient['dateModified'], pattern) > datetime.strptime(docClient['dateSync'], pattern) and \
       datetime.strptime(docOther['dateModified'], pattern)  > datetime.strptime(docOther['dateSync'], pattern):
      mergeCase = 5
      if self.verbose:
        print(f'** MERGE CASE {MERGE_LABELS[mergeCase]}')
      flagUpdateServer = True
      flagUpdateClient = False
      docMerged = copy.deepcopy(docClient)
    if mergeCase<=0:
      print(f'**ERROR** No merge case set! {mode}')
      return node.id, -1

    docMerged['dateSync'] = datetime.now().isoformat()
    if self.verbose:
      print('>>>MERGE TIME', docMerged['dateSync'])
    # update client
    if flagUpdateClient:
      docUpdate = copy.deepcopy(docMerged)
      docUpdate['branch'] = docUpdate['branch'][0] | {'op':''}
      docUpdate['dateModified'] = docUpdate['dateSync']
      if 'metaVendor' in docUpdate:
        del docUpdate['metaVendor']
      squashTupleIntoValue(docUpdate)
      self.backend.db.updateDoc(docUpdate, node.id)
    else:
      self.backend.db.cursor.execute(f"UPDATE main SET dateSync='{docMerged['dateSync']}' WHERE id = '{node.id}'")
      self.backend.db.connection.commit()
    # send doc (merged version) to server
    if flagUpdateServer:
      content, image = self.doc2elab(copy.deepcopy(docMerged))
      success = self.api.updateEntry(entryType, elabID, content|self.readWriteAccess)
      if not success:
        logging.error('Could not sync data %s, %s  %s',entryType, elabID, json.dumps(content|self.readWriteAccess, indent=2))
        return node.id, -1
      # create links
      _ = [self.api.createLink(entryType, elabID,
                               'experiments' if self.docID2elabID[i.id][1] else 'items', self.docID2elabID[i.id][0])
                               for i in node.children]
    # uploads| clean first, then upload: PASTAs document, thumbnail, data-file
    existingUploads = self.api.readEntry(entryType, elabID)[0]['uploads']
    uploadsToDelete = {'do_not_change.json', 'metadata.json'}
    if flagUpdateServer:
      uploadsToDelete |= {'thumbnail.svg', 'thumbnail.png', 'thumbnail.jpg'}
    for upload in existingUploads:
      if upload['real_name'] in uploadsToDelete:
        self.api.uploadDelete(entryType, elabID, upload['id'])
    self.api.upload(entryType, elabID, jsonContent=json.dumps(docMerged))
    if flagUpdateServer:
      if image:
        self.api.upload(entryType, elabID, image)
      if docMerged['branch'][0]['path'] is not None and docMerged['type'][0][0]!='x' \
            and not docMerged['branch'][0]['path'].startswith('http') and \
            (self.backend.basePath/docMerged['branch'][0]['path']).name not in {i['real_name'] for i in existingUploads}:
        self.api.upload(entryType, elabID, fileName=self.backend.basePath/docMerged['branch'][0]['path'], comment='raw data')
    return node.id, mergeCase


  def syncMissingEntries(self, mode:str='', callback:Callable[[ElabFTWApi,str,int],str]=cliCallback,
                         progressCallback:Callable[...,None]|None=None) -> list[tuple[str,int]]:
    """
    Compare information on server and client and delete those on server, that are not on client (because they were deleted there)

    Args:
      mode (str): sync mode g=get, gA=get-all, s=send, sA=send-all
      callback (func): callback function if non-all mode is given
      progressCallback (func): callback function to allow to monitor the progress

    Returns:
      list: changes
    """
    report = []
    self.backend.db.cursor.execute('SELECT type, externalId FROM main')
    dataLocal = self.backend.db.cursor.fetchall()
    inPasta = {'experiments': {int(i[1]) for i in dataLocal if i[0].startswith('measurement')},
               'items':       {int(i[1]) for i in dataLocal if not i[0].startswith('measurement')} }
    data = self.api.readEntry('items', self.elabProjGroupID)[0]
    inELAB  = {'experiments': {i['entityid'] for i in data['related_experiments_links']},
               'items':       {i['entityid'] for i in data['related_items_links']} }
    for count0, entryType in enumerate(['experiments','items']):
      if diff := inPasta[entryType].difference(inELAB[entryType]):
        print(f'**ERROR** There is a difference in {entryType} between CLIENT and SERVER. Ids on server: {diff}.')
      if diff := inELAB[entryType].difference(inPasta[entryType]):
        if self.verbose:
          print(f'**INFO** There is a difference in {entryType} between SERVER and CLIENT. Ids on server: {diff}.')
        for count1, idx in enumerate(diff):
          if progressCallback is not None:
            progressCallback('count', str(round(count0*50+count1*50/len(diff))))
          mode = mode if mode.endswith('A') else callback(self.api, entryType, idx)
          if mode.startswith('s'):
            self.api.deleteEntry(entryType, idx)
          elif mode.startswith('g'):
            _, uploads = self.elab2doc(self.api.readEntry(entryType, idx)[0])
            if listDoNotChange := [i for i in uploads if i['real_name']=='do_not_change.json']:
              docOther = self.api.download(entryType, idx, listDoNotChange[0])
              docOther['dateSync'] = datetime.now().isoformat()
              squashTupleIntoValue(docOther)
              docOther = flatten(docOther, keepPastaStruct=True)                       #type: ignore[assignment]
              try:
                branch = copy.deepcopy(docOther['branch'][0])
                # create folder
                if branch['path'] is not None:
                  if docOther['type'][0][0]=='x':
                    (self.backend.basePath/branch['path']).mkdir(parents=True, exist_ok=True)
                    with open(self.backend.basePath/branch['path']/'.id_pastaELN.json', 'w', encoding='utf-8') as fOut:
                      json.dump({'id':docOther['id']}, fOut)
                  else:
                    (self.backend.basePath/branch['path']).parent.mkdir(parents=True, exist_ok=True)
                docOther['branch'] = branch | {'op':'c'} #TODO: one remains only
                if 'tags' not in docOther:
                  docOther['tags'] = []
                self.backend.db.saveDoc(docOther)
                # save datafile if exists
                if listFile := [i for i in uploads if i['real_name']!='do_not_change.json' and \
                                not i['real_name'].startswith('thumbnail.')]:
                  data = self.api.download(listFile[0]['type'], idx, listFile[0])
                  with open(self.backend.basePath/branch['path'], 'wb') as fOut:
                    fOut.write(data['data'])
                report.append((docOther['id'], 2))
              except Exception:
                docOther.pop('image','')
                for k in list(docOther.keys()):
                  if k.startswith('metaVendor.'):
                    docOther.pop(k,'')
                print(f'**ERROR** Tried to add to client elab:{entryType} {idx}: {json.dumps(docOther,indent=2)}')
                print(traceback.format_exc())
                report.append((docOther['id'], -1))
    return report


  def elab2doc(self,elab:dict[str,Any]) -> tuple[dict[str,Any], list[Any]]:
    """ Convert an elabFTW dict-html entry into a PASTA doc

    Args:
      elab (dict): elabFTW entry

    Returns:
      dict: PASTA doc
    """
    # print(json.dumps(elab, indent=2))
    comment = html2markdown(elab.get('body','')).strip()
    # User is not allowed to change dates: ignore these dates from the server
    tags = [] if elab.get('tags','') is None else elab.get('tags','').split('|')
    doc = {'name': elab.get('title',''), 'tags':tags, 'comment':comment}
    metadata = {} if elab.get('metadata') is None else json.loads(elab['metadata'])
    # doc['metaVendor'] = metadata.get('metaVendor',{})  # USERS IS NOT ALLOWED TO CHANGE THESE
    # doc['metaUser']   = metadata.get('metaUser',{})
    doc |= metadata.get('__',{})                         # USERS CAN CHANGE THIS ON ELAB
    return doc, elab.get('uploads',[])


  def doc2elab(self, doc:dict[str,Any]) -> tuple[dict[str,Any], str]:
    """ Convert a PASTA doc into an elabFTW dict-html entry

    Args:
      doc (dict): PASTA doc

    Returns:
      dict: elabFTW entry
    """
    if not doc:
      raise ValueError('Cannot convert / process empty document')
    image     = doc.pop('image') if 'image' in doc else ''
    title     = doc.pop('name')
    bodyMD    = ''
    if 'content' in doc:
      bodyMD += doc.pop('content')+'\n\n__DO NOT CHANGE THIS DELIMITER between content (top) and comment (bottom)__\n\n'
    bodyMD    = doc.pop('comment')
    body      = markdown2html(bodyMD)
    # created_at= doc.pop('dateCreated')
    # modified_at=doc.pop('dateModified')
    tags       =doc.pop('tags')
    ratings    = [i[1] for i in tags if i.startswith('_')]
    tags       = [i for i in tags if not re.match(r'^_\d$', i)]
    doc        = {k:v for k,v in doc.items() if k not in {'id','user','client','externalId','type','gui','branch','shasum','qrCodes'}}
    metadata = {'metaVendor': doc.pop('metaVendor') if 'metaVendor' in doc else {}}
    metadata['metaUser']    = doc.pop('metaUser')   if 'metaUser' in doc else {}
    if doc:
      doc.pop('dateSync')
      metadata |= {'__':doc}
    elab = {'body':body, 'title':title, 'metadata':json.dumps(metadata), 'tags':tags,
            # 'created_at':created_at, 'modified_at':modified_at,
            'rating': ratings[0] if ratings else '0'}
    return elab, image



"""Parent class to repository classes"""
from typing import Any


class RepositoryClient:
  """Parent class to repository classes"""
  def __init__(self, server_url: str, api_token: str) -> None:
    """
    Initializes the client.

    Args:
        server_url (str): The URL of the server.
        api_token (str): The API token for authentication.
    """
    self.api_token = api_token
    self.server_url = server_url


  def checkServer(self) -> tuple[bool, str]:
    """
    Checks if the data-verse server is reachable

    Returns (tuple(bool, Any)):
      A tuple of (success, a message) is returned
    """
    return False, ''


  def checkAPIKey(self) -> bool:
    """
    Checks if the given API token is valid.

    Explanation:
        This method checks if the provided API token is valid by making a request to the server.
        It logs the server URL and sends a GET request to the token endpoint with the API token.
        It returns True if the response is successful and the status code is not 401, 403, or 500.

    Args:
        self: The instance of the class.

    Returns:
        bool: True if the API token is valid, False otherwise.
    """
    return False


  def uploadRepository(self, metadata:dict[str,Any], file_path:str) -> tuple[bool, str]:
    """
    Uploads a file and metadata to become a dataset.

    Args:
      metadata (dict): metadata to this file according to its standard
      file_path (str): The absolute path to the file to be uploaded.

    Returns:
      tuple: success of function, message
    """
    return False, f"Mock output of metadata {metadata} and file_path {file_path}"


  def prepareMetadata(self, metadata:dict[str,Any]) -> dict[str,Any]:
    """
    Prepares the metadata for uploading.

    Args:
        metadata (dict): The metadata to be prepared.

    Returns:
        dict: The prepared metadata.
    """
    return {}



""" all styling of buttons and other general widgets, some defined colors... """
import logging
import re
from typing import Any, Callable, Optional, Union
import qtawesome as qta
from PySide6.QtCore import QByteArray, Qt  # pylint: disable=no-name-in-module
from PySide6.QtGui import QAction, QImage, QKeySequence, QMouseEvent, QPixmap  # pylint: disable=no-name-in-module
from PySide6.QtSvgWidgets import QSvgWidget  # pylint: disable=no-name-in-module
from PySide6.QtWidgets import (QApplication, QBoxLayout, QComboBox, QDialog,  # pylint: disable=no-name-in-module
                               QFormLayout, QGridLayout, QHBoxLayout, QLabel, QLayout, QMenu, QMessageBox, QPushButton,
                               QScrollArea, QSizePolicy, QSplitter, QTextEdit, QVBoxLayout, QWidget)
from .textTools.handleDictionaries import dict2ul
from .textTools.stringChanges import markdownEqualizer

space = {'0':0, 's':5, 'm':10, 'l':20, 'xl':200} #spaces: padding and margin

class TextButton(QPushButton):
  """ Button that has only text"""
  def __init__(self, label:str, widget:QWidget, command:list[Any]|None=[], layout:Optional[QLayout]=None,
               tooltip:str='', checkable:bool=False, style:str='', hide:bool=False, iconName:str=''):
    """
    Args:
      label (str): label printed on button
      widget (QWidget): widget / dialog that host the button and that has the execute function
      command (enum): command that is used in called-function: possibly a list of multiple terms
      layout (QLayout): button to be added to this layout
      tooltip (str): tooltip shown when mouse hovers the button
      checkable (bool): can the button change its background color
      style (str): css style
      hide (bool): hidden or shown initially
    """
    super().__init__()
    self.setText(label)
    self.setCheckable(checkable)
    self.setChecked(checkable)
    self.setAutoDefault(False)
    self.setDefault(False)
    if command is not None:
      self.clicked.connect(lambda: widget.execute(command))                                             # type: ignore[attr-defined]
    if tooltip:
      self.setToolTip(tooltip)
    if style:
      self.setStyleSheet(style)
    else:
      primaryColor   = widget.comm.palette.get('primary',  'background-color')                          # type: ignore[attr-defined]
      secondaryColor = widget.comm.palette.get('buttonText','color')                                    # type: ignore[attr-defined]
      self.setStyleSheet(f'border-width: 0px; {primaryColor} {secondaryColor}')
    if hide:
      self.hide()
    if iconName:
      icon = qta.icon(iconName, scale_factor=1)
      self.setIcon(icon)
    if layout is not None:
      layout.addWidget(self)


class IconButton(QPushButton):
  """ Button that has only an icon"""
  def __init__(self, iconName:str, widget:QWidget, command:list[Any]=[], layout:Optional[QLayout]=None,
               tooltip:str='', style:str='', hide:bool=False, checkable:bool=False):
    """
    Args:
      iconName (str): icon to show on button
      widget (QWidget): widget / dialog that host the button and that has the execute function
      command (enum): command that is used in called-function: possibly a list of multiple terms
      layout (QLayout): button to be added to this layout
      tooltip (str): tooltip shown when mouse hovers the button
      style (str): css style
      hide (bool): hidden or shown initially
      checkable (bool): can the button change its background color
    """
    super().__init__()
    icon = qta.icon(iconName, scale_factor=1)  #color change here
    self.setIcon(icon)
    self.setCheckable(checkable)
    self.command = command
    self.clicked.connect(lambda: widget.execute(self.command))                                                    # type: ignore[attr-defined]
    self.setFixedHeight(30)
    if tooltip:
      self.setToolTip(tooltip)
    if style:
      self.setStyleSheet(f'QPushButton {{{style}}} QPushButton:checked {{border: "red"; border-width: 5px; {style}}}')
    else:
      primaryColor   = widget.comm.palette.get('primary', 'background-color')                          # type: ignore[attr-defined]
      secondaryColor = widget.comm.palette.get('secondary','color')                                    # type: ignore[attr-defined]
      self.setStyleSheet(f'QPushButton {{border-width: 0px; {primaryColor} {secondaryColor}}} QPushButton:checked {{border: 3px solid red; {primaryColor} {secondaryColor}}}')
    if hide:
      self.hide()
    if layout is not None:
      layout.addWidget(self)


class Action(QAction):
  """ QAction and assign function to menu"""
  def __init__(self, label:str, widget:QWidget, command:list[Any],
               menu:QMenu, shortcut:Optional[str]=None, icon:str=''):
    """
    Args:
      label (str): label printed on submenu
      widget (QWidget): widget / dialog that host the button and that has the execute function
      command (enum): command that is used in called-function: possibly a list of multiple terms
      menu (QMenu): button to be added to this menu
      shortcut (str): shortcut (e.g. Ctrl+K)
      icon (str): icon name
    """
    super().__init__()
    self.setParent(widget)
    self.setText(label)
    self.triggered.connect(lambda : widget.execute(command))                                    # type: ignore[attr-defined]
    if icon:
      color = 'black' if widget is None else widget.comm.palette.secondaryText                  # type: ignore[attr-defined]
      self.setIcon(qta.icon(icon, color=color, scale_factor=1))
    if shortcut is not None:
      self.setShortcut(QKeySequence(shortcut))
    menu.addAction(self)


class Image():
  """ Image widget depending on type of data """
  def __init__(self, data:str, layout:Optional[QLayout], width:int=-1, height:int=-1, anyDimension:int=-1):
    """
    Args:
      data (str): image data in byte64-encoding or svg-encoding
      layout (QLayout): to be added to this layout
      width (int): width of image, dominant if both are given
      height (int): height of image
      anyDimension (int): maximum size in any direction
    """
    if data.startswith('data:image/'): #jpg or png image
      byteArr = QByteArray.fromBase64(bytearray(data[22:] if data[21]==',' else data[23:], encoding='utf-8'))
      imageW = QImage()
      imageType = data[11:15].upper()
      imageW.loadFromData(byteArr, format=imageType[:-1] if imageType.endswith(';') else imageType) #type: ignore[arg-type]
      pixmap = QPixmap.fromImage(imageW)
      if height>0:
        pixmap = pixmap.scaledToHeight(height)
      if width>0:
        pixmap = pixmap.scaledToWidth(width)
      if anyDimension>0:
        if pixmap.size().height()>pixmap.size().width():
          pixmap = pixmap.scaledToHeight(anyDimension)
        else:
          pixmap = pixmap.scaledToWidth(anyDimension)
      label = QLabel()
      label.setPixmap(pixmap)
      label.setAlignment(Qt.AlignCenter) # type: ignore
      if layout is not None:
        layout.addWidget(label, alignment=Qt.AlignHCenter)  # type: ignore
    elif data.startswith('<?xml'): #svg image
      imageSVG = QSvgWidget()
      policy = imageSVG.sizePolicy()
      policy.setHorizontalPolicy(QSizePolicy.Policy.Fixed)
      policy.setVerticalPolicy(QSizePolicy.Policy.Fixed)
      imageSVG.setSizePolicy(policy)
      imageSVG.renderer().load(bytearray(data, encoding='utf-8'))
      if height>0:
        imageSVG.setMaximumSize(int(float(imageSVG.width())/float(imageSVG.height())*height) ,height)
      if width>0:
        imageSVG.setMaximumSize(width, int(float(imageSVG.height())/float(imageSVG.width())*width))
      if anyDimension>0:
        if imageSVG.height()>imageSVG.width():
          imageSVG.setMaximumSize(int(float(imageSVG.width())/float(imageSVG.height())*anyDimension) ,anyDimension)
        else:
          imageSVG.setMaximumSize(anyDimension, int(float(imageSVG.height())/float(imageSVG.width())*anyDimension))
      if layout is not None:
        layout.addWidget(imageSVG, alignment=Qt.AlignHCenter) # type: ignore
    elif len(data)>2:
      print(f'WidgetProjectLeaf:What is this image |{data[:50]}|')
    return


class Label(QLabel):
  """ Label widget: headline, ... """
  def __init__(self, text:str='', size:str='', layout:Optional[QLayout]=None,
               function:Optional[Callable[[str, str],None]]=None, docID:str='', tooltip:str='', style:str=''):
    """ Label widget with given font-size and functions:
    - text selection: if only character, easy selection
      - if formatted text: right-mouse-button to select all (There is no other way, apparently)

    Args:
      text (str): text on label
      size (str): size ['h1','h2','h3']
      layout (QLayout): layout to which to add the label
      function (function): function to call on mouse click
      docID (str): docID on other string to connect to this label
      tooltip (str): tooltip shown when mouse hovers the button
      style (str): css style
    """
    super().__init__()
    self.setText(text)
    if text.startswith('#') or text.startswith('<'):
      self.setTextFormat(Qt.TextFormat.RichText)
    self.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse | Qt.TextInteractionFlag.LinksAccessibleByMouse)
    if size == 'h1':
      style += 'font-size: 14pt'
    elif size == 'h2':
      style += 'font-size: 12pt'
    elif size == 'h3':
      style += 'font-size: 10pt'
    if style:
      self.setStyleSheet(style)
    if layout is not None:
      layout.addWidget(self)
    self.mouseFunction = function
    self.identifier = docID
    if tooltip != '':
      self.setToolTip(tooltip)
    return

  def mousePressEvent(self, _:QMouseEvent) -> None:
    """
    Event after mouse press: only use internal members, not the event itself
    """
    if self.mouseFunction is not None:
      self.mouseFunction(self.text(), self.identifier)
    return



def showMessage(parent:QWidget, title:str, text:str, icon:str='Information', style:str='') -> None:
  """
  Show simple message box for little information

  Args:
    parent (QWidget): parent widget (self)
    title (str): title of box
    text (str): text in box
    icon (str): icon: 'Information','Warning','Critical'
    style (str): css style
  """
  iconSize = 40
  color = 'red' if icon=='Critical' else '#ffbc00' if icon=='Warning' else '#'
  iconSymbol = qta.icon('fa5s.minus-circle' if icon=='Critical' else
                  'fa5s.exclamation-circle' if icon=='Warning' else
                  'fa5s.info', color='white', scale_factor=1)
  dialog = QDialog(parent)
  dialog.setWindowTitle(title)
  dialogW, dialogL = widgetAndLayout('V', None, 's', top='s', bottom='s', left='s', right='s')
  dialog.setLayout(dialogL)
  label = QLabel('')
  label.setPixmap(iconSymbol.pixmap(iconSize, iconSize))
  label.setStyleSheet(f'background: {color};')
  label.setMinimumSize(iconSize, iconSize)
  dialogL.addWidget(label, alignment=Qt.AlignHCenter)  # type: ignore
  label = Label(text, 'h2', dialogL)
  label.setStyleSheet(style)
  if text.startswith('<') and text.endswith('>'):
    label.setTextFormat(Qt.TextFormat.RichText)
    text = text.replace('<font color="black">',f'<font color="{parent.comm.palette.get("secondaryText", "").strip()[2:-1]}">') # type: ignore[attr-defined]
    label.setText(text)
  else:
    label.setTextFormat(Qt.TextFormat.MarkdownText)
  # final button box
  _, buttonLineL = widgetAndLayout('H', dialogL, 'm', 'm', '0', 'm', 's')
  copyButton = TextButton('Copy message', parent, None, buttonLineL, 'Copy to clipboard')
  def copyToClipboard() -> None:
    """ Copy text to clipboard """
    clipboard = QApplication.clipboard()
    clipboard.setText(text)
  copyButton.clicked.connect(copyToClipboard)
  buttonLineL.addStretch(2)
  okButton   = TextButton('OK',   parent, None, buttonLineL, 'Accept')
  okButton.clicked.connect(dialog.accept)
  dialogW.setMinimumWidth(800)
  dialog.exec()
  return


class ScrollMessageBox(QMessageBox):
  """ Scrollable message box for lots of dictionary information """
  def __init__(self, title:str, content:dict[str,Any], style:str=''):
    """
    Args:
      title (str): title
      content (dict): dictionary of lots of information
      style (str): css style
    """
    cssStyle = '<style> ul {list-style-type: none; padding-left: 0; margin: 0; text-indent: -20px; padding-left: -20px;} </style>'
    QMessageBox.__init__(self)
    self.setWindowTitle(title)
    if not style:
      self.setStyleSheet('QScrollArea{min-width:300 px; min-height: 400px}')
    else:
      self.setStyleSheet(style)
    scroll = QScrollArea(self)
    scroll.setWidgetResizable(True)
    self.content = QLabel()
    self.content.setWordWrap(True)
    self.content.setText(cssStyle+dict2ul(content))
    self.content.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)
    scroll.setWidget(self.content)
    self.layout().addWidget(scroll, 0, 0, 1, self.layout().columnCount()) # type: ignore


def widgetAndLayout(direction:str='V', parentLayout:Optional[Union[QLayout,QSplitter]]=None, spacing:str='0', left:str='0',
                    top:str='0', right:str='0', bottom:str='0') -> tuple[QWidget, QBoxLayout]:
  """
  Convenient function for widget and a boxLayout

  Spacings and margins:
  - different than in css/html
  - spacing is the space between elements in the orientation of the BoxLayout
  - is the padding that surrounds the content in the layout

  Distances are given in
  - '0': zero distance
  - 's': small distance used as padding round elements, or vertical spacings
  - 'm': medium used as space between horizontal elements
  - 'l': large used when things need to be separated
  - 'xl': extra large indentations, frames

  Args:
    direction (str): type of layout [H,V]
    parentLayout (QLayout): to which layout should the widget be added. If none, no adding
    spacing (str): spacing
    left (str): padding on left
    top (str): padding on top
    right (str): padding on right
    bottom (str): padding on bottom
  """
  widget = QWidget()
  layout = QVBoxLayout(widget) if direction=='V' else QHBoxLayout(widget)
  layout.setSpacing(space[spacing])
  layout.setContentsMargins(space[left], space[top], space[right], space[bottom])
  if parentLayout is not None:
    parentLayout.addWidget(widget)
  return widget, layout


def widgetAndLayoutForm(parentLayout:Optional[Union[QLayout,QSplitter]]=None, spacing:str='0', left:str='0',
                    top:str='0', right:str='0', bottom:str='0') -> tuple[QWidget, QFormLayout]:
  """
  Convenient function for widget and a form layout
  - comment see above

  Args:
    parentLayout (QLayout): to which layout should the widget be added. If none, no adding
    spacing (str): spacing
    left (str): padding on left
    top (str): padding on top
    right (str): padding on right
    bottom (str): padding on bottom
  """
  widget = QWidget()
  layout = QFormLayout(widget)
  layout.setSpacing(space[spacing])
  layout.setContentsMargins(space[left], space[top], space[right], space[bottom])
  if parentLayout is not None:
    parentLayout.addWidget(widget)
  return widget, layout


def widgetAndLayoutGrid(parentLayout:Optional[QLayout]=None, spacing:str='0', left:str='0',
                    top:str='0', right:str='0', bottom:str='0') -> tuple[QWidget, QGridLayout]:
  """
  Convenient function for widget and a grid layout
  - comment see above

  Args:
    parentLayout (QLayout): to which layout should the widget be added. If none, no adding
    spacing (str): spacing
    left (str): padding on left
    top (str): padding on top
    right (str): padding on right
    bottom (str): padding on bottom
  """
  widget = QWidget()
  layout = QGridLayout(widget)
  layout.setSpacing(space[spacing])
  layout.setContentsMargins(space[left], space[top], space[right], space[bottom])
  if parentLayout is not None:
    parentLayout.addWidget(widget)
  return widget, layout


def addRowList(layout:QFormLayout, label:str, default:str, itemList:list[str]) -> QComboBox:
  """
  Add a row with a combo-box to the form

  Args:
    layout (QLayout): layout to add row to
    label (str): label used in form
    default (str): default value
    itemList (list(str)): items to choose from

  Returns:
    QCombobox: filled combobox
  """
  widget = QComboBox()
  widget.addItems(itemList)
  widget.setCurrentText(default)
  layout.addRow(QLabel(label), widget)
  return widget

CSS_STYLE = """
<style> ul {list-style-type: none; padding-left: 0; margin: 0;} a:link {text-decoration: none;}
a:visited {text-decoration: none;} a:hover {text-decoration: none;} a:active {text-decoration: none;} </style>
"""

def addDocDetails(widget:QWidget, layout:QLayout, key:str, value:Any, dataHierarchyNode:list[dict[str,Any]]) -> str:
  """ add document details to a widget's layout: take care of all the formatting

  Args:
    widget (QWidget): widget from which to get the communication
    layout (QLayout): layout to which to add
    key    (str): key/label to add
    value  (Any): information
    dataHierarchyNode (list): information on data-structure

  Returns:
    str: /n separated lines of text
  """
  if not key and isinstance(value,dict):
    return '\n'.join([addDocDetails(widget, layout, k, v, dataHierarchyNode) for k, v in value.items()])
  link = False
  if not value:
    return ''
  labelStr = ''
  if key=='tags':
    rating = ['\u2605'*int(i[1]) for i in value if re.match(r'^_\d$', i)]
    tags = ['_curated_' if i=='_curated' else i for i in value]
    tags = [i for i in tags if not re.match(r'^_\d$', i)]
    labelStr = f'Rating: {rating[0]}' if rating else ''
    labelStr = f'{labelStr}   Tags: '+' '.join(tags)
    if layout is not None:
      label = QLabel(labelStr)
      label.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)
      layout.addWidget(label)
  elif (isinstance(value,str) and '\n' in value) or key=='comment':                 # long values with /s or comments
    labelW, labelL = widgetAndLayout('H', layout, top='s', bottom='s')
    labelL.addWidget(QLabel(f'{key}: '), alignment=Qt.AlignmentFlag.AlignTop)
    text = QTextEdit()
    text.setMarkdown(markdownEqualizer(value))
    bgColor = widget.comm.palette.get('secondaryDark', 'background-color') # type: ignore[attr-defined]
    fgColor = widget.comm.palette.get('secondaryText', 'color')             # type: ignore[attr-defined]
    text.setStyleSheet(f"QTextEdit {{ border: none; padding: 0px; {bgColor} {fgColor}}}")
    text.document().setTextWidth(labelW.width())
    if hasattr(widget, 'rescaleTexts'):
      widget.rescaleTexts.append(text)
    height:int = text.document().size().toTuple()[1]    # type:ignore[index]
    text.setFixedHeight(height)
    text.setReadOnly(True)
    labelL.addWidget(text, stretch=1)
  else:
    dataHierarchyItems = [dict(i) for i in dataHierarchyNode if i['name']==key]
    docID = ''
    if len(dataHierarchyItems)==1 and 'list' in dataHierarchyItems[0] and dataHierarchyItems[0]['list'] and \
        not isinstance(dataHierarchyItems[0]['list'], list):                #choice among docType
      table  = widget.comm.backend.db.getView('viewDocType/'+dataHierarchyItems[0]['list']) # type: ignore[attr-defined]
      names= list(table[table.id==value[0]]['name'])
      if len(names)==1:    # default find one item that we link to
        docID = value[0]
        value = '\u260D '+names[0]
        link = True
      elif not names:      # likely empty link because the value was not yet defined: just print to show
        value = value[0] if isinstance(value,tuple) else value
      else:
        raise ValueError(f'list target exists multiple times. Key: {key}')
    elif isinstance(value, list):
      value = ', '.join([str(i) for i in value])
    labelStr = f'<b>{key.capitalize()}</b>: {value}'
    if isinstance(value, tuple) and len(value)==4:
      key = key if value[2] is None or value[2]=='' else value[2]
      valueString = f'{value[0]} {value[1]}'
      valueString = valueString if value[3] is None or value[3]=='' else f'{valueString}&nbsp;<b><a href="{value[3]}">&uArr;</a></b>'
      labelStr = f'{key.capitalize()}: {valueString}<br>'
    if isinstance(value, dict):
      value = {k:(v[0] if isinstance(v, (list,tuple)) else v) for k,v in value.items()}
      labelStr = f'{CSS_STYLE}{key.capitalize()}: {dict2ul(value)}'
    if layout is not None:
      label = Label(labelStr, function=lambda x,y: clickLink(widget,x,y) if link else None, docID=docID)
      label.setOpenExternalLinks(True)
      label.setWordWrap(True)
      layout.addWidget(label)
  return labelStr


def clickLink(widget:QWidget, label:str, docID:str) -> None:
  """
  Click link in details

  Args:
    widget (QWidget): widget to expose the communication
    label (str): label on link
    docID (str): docID to which to link
  """
  logging.debug('used link on %s|%s',label,docID)
  widget.comm.changeDetails.emit(docID)   # type: ignore[attr-defined]
  return



from pasta_eln.gui import MainWindow

def test_simple(qtbot):
  window = MainWindow('research')
  window.setMinimumSize(1024,800)
  window.show()
  qtbot.addWidget(window)

  # projID1 = window.comm.backend.output('x0').split('|')[-1].strip()
  # print(projID1)
  # window.comm.changeProject.emit(projID1, '')
  # click in the Greet button and make sure it updates the appropriate label
  # projectBtn = window.sidebar.projectsListL.itemAt(0).widget().layout().itemAt(0).widget()
  # qtbot.mouseClick(projectBtn, Qt.LeftButton)
  #
  path = qtbot.screenshot(window)
  print(path)
  # saved in
  #   /tmp/pytest-of-steffen/pytest-0/test_simple0/
  #   /tmp/pytest-of-runner/pytest-0/test_simple0/screenshot_MainWindow.png



"""example addon: plot data from files

THIS IS A VERY ADVANCED ADDON TUTORIAL
This tutorial teaches
- how to plot the data from files
"""
import matplotlib.pyplot as plt
import pandas as pd
from pasta_eln.miscTools import callDataExtractor

# The following two variables are mandatory
description  = 'Default data plot'  #short description that is shown in the menu
reqParameter = {} #possibility for required parameters: like API-key, etc. {'API': 'text'}

def main(backend, df, widget, parameter={}):
    """ main function: has to exist and is called by the menu
    Args:
        backend (pasta backend): allow to extract data
        df (Dataframe): dataframe with the data to plot. The docIDand path columns are of most interest
        widget (QWidget): allows to create new gui dialogs
        parameter (dict): ability to pass parameters

    Returns:
        bool: success
    """
    # no cleaning needed since we only use id and path columns
    for idx, row in df.iterrows():
        if not row['path'].endswith('.csv'):          #by default, only the csv file extractor has a data function; you can change that
            continue
        res = callDataExtractor(row['path'], backend)
        if res is None:
            continue
        # Here you can process the extracted data further, e.g., plot it
        if isinstance(res, pd.DataFrame):
            res.plot()
            plt.show()
        else:
            print(f"Extracted data is not a DataFrame. {row['path']}")
    return True



""" Communication class that sends signals between widgets, incl. backend"""
from typing import Any, Callable
from PySide6.QtCore import QObject, Signal  # pylint: disable=no-name-in-module
from .backend import Backend
from .GUI.waitDialog import WaitDialog, Worker


class Communicate(QObject):
  """ Communication class that sends signals between widgets, incl. backend"""
  def __init__(self, backend:Backend, palette:Any):
    super().__init__()
    self.backend               = backend
    self.palette               = palette
    self.projectID             = ''
    self.waitDialog            = WaitDialog()
    self.worker:Worker|None    = None

  def progressWindow(self, taskFunction:Callable[[Callable[[str,str],None]],Any]) -> None:
    """ Show a progress window and execute function
    Args:
      taskFunction (func): function to execute
    """
    self.waitDialog.show()
    self.worker = Worker(taskFunction)
    self.worker.progress.connect(self.waitDialog.updateProgressBar)
    self.worker.start()
    return



  # Signals: specify emitter and receiver
  # BE SPECIFIC ABOUT WHAT THIS ACTION DOES
  changeSidebar      = Signal(str)       # redraw sidebar after hide/show of project in table, focus on this projectID
  changeTable        = Signal(str, str)  # send doctype,projectID from sidebar to main-table
                                         #      can also be used for hiding the details on right side if nothing to show
  changeDetails      = Signal(str)       # send docID from main-table to details
                                         #      docID (str): document-id; ''=draw nothing; 'redraw' implies redraw
  changeProject      = Signal(str, str)  # send docID,projectID from sidebar or main-table to projects
  formDoc            = Signal(dict)      # send doc from details to new/edit dialog: dialogForm
  testExtractor      = Signal()          # execute extractorTest in widgetDetails
  stopSequentialEdit = Signal()          # in sequential edit, stop if there is a cancel
  restart            = Signal()          # restart GUI  #TODO rename to softRestart to avoid confusion



""" Config for Zenodo and Dataverse """
import json
import re
import webbrowser
from enum import Enum
from pathlib import Path
from typing import Any, Callable
import qtawesome as qta
from PySide6.QtWidgets import QComboBox, QDialog, QLabel, QLineEdit, QVBoxLayout  # pylint: disable=no-name-in-module
from ...fixedStringsJson import CONF_FILE_NAME
from ...guiCommunicate import Communicate
from ...guiStyle import Label, TextButton, showMessage, widgetAndLayout, widgetAndLayoutGrid
from .dataverse import DataverseClient
from .zenodo import ZenodoClient


class ConfigurationRepositories(QDialog):
  """ Config for Zenodo and Dataverse """
  def __init__(self, comm:Communicate, callbackFinished:Callable[[bool],None]):
    """
    Initialization

    Args:
      backend (Backend): PASTA-ELN backend
      callbackFinished (function): callback function to call upon end
    """
    super().__init__()
    self.comm    = comm
    self.callbackFinished = callbackFinished
    self.configuration = self.comm.backend.configuration
    self.checkedZenodo = False
    self.checkedDataverse = True
    self.setStyleSheet(f"QLineEdit {{ {self.comm.palette.get('secondaryText', 'color')} }}")

    # GUI elements
    mainL = QVBoxLayout(self)
    Label('Configure the repositories', 'h1', mainL)
    _, center = widgetAndLayout('H', mainL, spacing='l', bottom='l', top='m')

    leftSideW, leftSide = widgetAndLayoutGrid(center, spacing='m', right='l')
    leftSideW.setStyleSheet('border-right: 2px solid black;')
    leftSide.addWidget(QLabel('Zenodo'), 0, 0)
    leftSide.addWidget(QLabel('URL'), 1, 0)
    self.urlZenodo = QLineEdit('https://zenodo.org')
    self.urlZenodo.setMinimumWidth(350)
    leftSide.addWidget(self.urlZenodo, 1, 1)
    leftSide.addWidget(QLabel('API key'), 2, 0)
    self.apiZenodo = QLineEdit()
    leftSide.addWidget(self.apiZenodo, 2, 1)
    self.zenodoButton = TextButton('Check',   self, [Command.CHECK_ZENODO], tooltip='Check Zenodo login details')
    leftSide.addWidget(self.zenodoButton, 3, 1)

    _, rightSide = widgetAndLayoutGrid(center, spacing='m', right='l')
    rightSide.addWidget(QLabel('Dataverse'), 0, 0)
    rightSide.addWidget(QLabel('URL'), 1, 0)
    self.urlDatavese = QLineEdit()
    self.urlDatavese.setMinimumWidth(350)
    rightSide.addWidget(self.urlDatavese, 1, 1)
    self.dataverseButton1 = TextButton('Check',   self, [Command.CHECK_DV1], tooltip='Check Dataverse server details')
    self.dataverseButton1.setMinimumWidth(100)
    rightSide.addWidget(self.dataverseButton1, 1, 2)
    rightSide.addWidget(QLabel('API key'), 2, 0)
    self.apiDataverse = QLineEdit()
    rightSide.addWidget(self.apiDataverse, 2, 1)
    self.dataverseButton2 = TextButton('Check',   self, [Command.CHECK_DV2], tooltip='Check Dataverse API-key')
    rightSide.addWidget(self.dataverseButton2, 2, 2)
    self.dataverseButton2.setMinimumWidth(100)
    rightSide.addWidget(QLabel('Sub dataverse'), 3, 0)
    self.dvDataverse = QComboBox()
    self.dvDataverse.setStyleSheet(self.comm.palette.get('secondaryText', 'color'))
    rightSide.addWidget(self.dvDataverse, 3, 1)

    #final button box
    _, buttonLineL = widgetAndLayout('H', mainL, 'm')
    TextButton('Help',           self, [Command.HELP],   buttonLineL, 'Help for this dialog')
    buttonLineL.addStretch(1)
    saveBtn = TextButton('Save', self, [Command.SAVE],   buttonLineL, 'Save changes')
    saveBtn.setShortcut('Ctrl+Return')
    TextButton('Cancel',         self, [Command.CANCEL], buttonLineL, 'Discard changes')


  def execute(self, command:list[Any]) -> None:
    """
    Event if user clicks button in the center

    Args:
      command (list): list of commands
    """
    if command[0] is Command.CHECK_ZENODO:
      url = self.urlZenodo.text().strip()
      if re.match(r'(http:|https:)+\/\/[\w\.]+', url) is None:
        showMessage(self, 'Error', 'URL is not valid')
        return
      api = self.apiZenodo.text().strip()
      if re.match(r'\w{60}', api) is None:
        showMessage(self, 'Error', 'API key is not valid')
        return
      clientZ = ZenodoClient(url, api)
      success, message = clientZ.checkServer()
      if success:
        self.changeButtonOnTest(True, self.zenodoButton)
        self.checkedZenodo = True
      else:
        self.changeButtonOnTest(False, self.zenodoButton)
        showMessage(self, 'Error', message)
    elif command[0] is Command.CHECK_DV1:
      url = self.urlDatavese.text().strip()
      if re.match(r'(http:|https:)+\/\/[\w\.]+', url) is None:
        showMessage(self, 'Error', 'URL is not valid')
        return
      clientD = DataverseClient(url, '', '')
      success, message = clientD.checkServer()
      if success:
        self.changeButtonOnTest(True, self.dataverseButton1)
      else:
        self.changeButtonOnTest(False, self.dataverseButton1)
        showMessage(self, 'Error', message)
    elif command[0] is Command.CHECK_DV2:
      url = self.urlDatavese.text().strip()
      if re.match(r'(http:|https:)+\/\/[\w\.]+', url) is None:
        showMessage(self, 'Error', 'URL is not valid')
        return
      api = self.apiDataverse.text().strip()
      if re.match(r'\w{8}-\w{4}-\w{4}-\w{4}-\w{12}', api) is None:
        showMessage(self, 'Error', 'API key is not valid')
        return
      clientD = DataverseClient(url, api, '')
      if success := clientD.checkAPIKey():
        self.changeButtonOnTest(True, self.dataverseButton2)
        self.checkedDataverse = True
        self.dvDataverse.clear()
        for data in clientD.getDataverseList():
          self.dvDataverse.addItem(f"{data.get('title')} - {data.get('id')}", data.get('id'))
      else:
        self.changeButtonOnTest(False, self.dataverseButton2)
        showMessage(self, 'Error', 'API key invalid')
    elif command[0] is Command.CANCEL:
      self.reject()
      self.callbackFinished(False)
    elif command[0] is Command.SAVE:
      if 'repositories' not in self.configuration:
        self.configuration['repositories'] = {}
      if self.checkedZenodo:
        self.configuration['repositories']['zenodo'] = {'url':self.urlZenodo.text(),
                                                        'key':self.apiZenodo.text()}
      if self.checkedDataverse:
        self.configuration['repositories']['dataverse'] = {'url':self.urlDatavese.text(),
                                                          'key':self.apiDataverse.text(),
                                                          'dataverse':self.dvDataverse.currentData()}
      with open(Path.home()/CONF_FILE_NAME, 'w', encoding='utf-8') as fConf:
        fConf.write(json.dumps(self.configuration,indent=2))
      self.accept()
    elif command[0] is Command.HELP:
      webbrowser.open('https://pasta-eln.github.io/pasta-eln/repositories.html')
    else:
      print('Got some button, without definition', command)
    return


  def changeButtonOnTest(self, success:bool, button:TextButton, message:str='') -> None:
    """ Helper function to change buttons upon success/failure

    Args:
      success (bool): change to successful state
      button (TextBotton): button to change
      message (str): optional message on failure
    """
    if success:
      button.setStyleSheet('background: #00FF00')
      button.setText('')
      button.setIcon(qta.icon('fa5s.check-square', scale_factor=1))
    else:
      if message:
        showMessage(self, 'Error', message)
      button.setStyleSheet('background: #FF0000')
      button.setText('')
      button.setIcon(qta.icon('fa5.times-circle', scale_factor=1))
    return


class Command(Enum):
  """ Commands used in this file """
  CHECK_ZENODO= 1
  CHECK_DV1   = 2
  CHECK_DV2   = 3
  HELP        = 4
  SAVE        = 5
  CANCEL      = 6



"""
Copied from https://github.com/Alir3z4/html2text
- which has a GPL-3.0 license
- simplified into four files, removed the library constraint
"""
from typing import Dict, Optional


class AnchorElement:
    __slots__ = ['attrs', 'count', 'outcount']

    def __init__(self, attrs: dict[str, Optional[str]], count: int, outcount: int):
        self.attrs = attrs
        self.count = count
        self.outcount = outcount


class ListElement:
    __slots__ = ['name', 'num']

    def __init__(self, name: str, num: int):
        self.name = name
        self.num = num



"""extract data from vendor
- default tif file
"""
import base64
from io import BytesIO
from PIL import Image


def use(filePath, style={'main':''}, saveFileName=None):
  """
  Args:
    filePath (Path): full path file name
    style    (dict): supplied to guide the display / extraction style = recipe
                     main is / separated hierarchical elements parent->child
                     can contain more elements
    saveFileName (string): if given, save the image to this file-name

  Returns:
    dict: containing image, metaVendor, metaUser, style
  """
  image = Image.open(filePath)
  metaVendor = image.info
  toDelete = [key for key, value in metaVendor.items() if not isinstance(value, (str, int, float, list))]
  for key in toDelete:
    del metaVendor[key]
  style['main'] = 'measurement/image'
  metaUser   = {}

  # save to file: this is the high quality image
  if saveFileName is not None:
    image.save(saveFileName)

  # convert PIL image to base64: here one should downscale
  maxSize = 400
  if max(image.size)>maxSize:
    scale = max(image.size)/maxSize
    newSize = (int(image.size[0]/scale), int(image.size[1]/scale))
    image = image.resize(newSize)

  image = image.convert('RGB')
  figfile = BytesIO()
  image.save(figfile, format='JPEG')
  image = base64.b64encode(figfile.getvalue()).decode()
  image = f"data:image/jpg;base64,{image}"

  # return everything
  return {'image':image, 'style':style, 'metaVendor':metaVendor, 'metaUser':metaUser}



""" Example workflow for the Sandia Fracture Challenge 3 """
# pylint: skip-file
# head of workflow: always the same
from urllib.parse import urlparse
from common_workflow_description.common_workflow_description import Sample, Storage, step

try:
  from pyiron_workflow import Workflow
except ImportError:
  from common_workflow_description.common_workflow_description import Workflow

from analysis_steps import calc_YoungsModulus, plot_curves

# start code
wf = Workflow('Sandia Fracture Challenge 3', automate_execution=False)         # name
proceduresLibrary = urlparse('https://raw.githubusercontent.com/SteffenBrinckmann/common-workflow-description_Procedures/main')
storage=Storage(proceduresLibrary)                                             # folder of database

# body of workflow: this changes
sample = Sample('AM_NA_05')

wf.step1 = step(storage, sample, 'metallography', {}, autorun=True)   #define step and link to storage for procedures
wf.step2 = step(storage, sample, 'light microscopy', {}, autorun=True)
wf.step3 = step(storage, sample, 'tensile test', {}, autorun=True)
wf.step4 = step(storage, sample, 'light microscopy', {}, autorun=True)
wf.step5a = step(storage, sample, 'sem', {'voltage':'30'}, autorun=True)
wf.step5b = step(storage, sample, 'sem', {'voltage':'30'}, autorun=True)

wf.step6 = plot_curves(wf.step3.outputs.y[1], 'Strain (Gauge0)', 'Engr. Stress')
wf.step7 = calc_YoungsModulus(wf.step3.outputs.y[1], 'Strain (Gauge0)', 'Engr. Stress')

wf.step1 >> wf.step2 >> wf.step3 >> wf.step4 >> wf.step5a >> wf.step5b >> wf.step6 >> wf.step7
wf.starting_nodes = [wf.step1]

# footer, always the same
print('Output:\n  ','\n   '.join([f"{k}: {v}" for k, v in list(wf.outputs.to_value_dict().items())]))
wf.draw().render(filename='io_demo_2', format='png', cleanup=True) #plot to file

# {"type":"common-workflow-description", "version":1.0, "shasum":"d5d8342b266851e28919e69039918d2c2a9ca4f2"}



from PySide6.QtWidgets import QMainWindow
from pasta_eln.backend import Backend
from pasta_eln.guiCommunicate import Communicate
from pasta_eln.GUI.palette import Palette
from pasta_eln.GUI.project import Project

def test_simple(qtbot):
  class MainWindow(QMainWindow):
    def __init__(self):
      super().__init__()
      self.backend = Backend('research')
      palette = Palette(self,'none')
      comm = Communicate(self.backend, palette)
      widget = Project(comm)
      self.setCentralWidget(widget)
      projID1 = self.backend.output('x0').split('|')[-1].strip()
      comm.changeProject.emit(projID1, '')

  window = MainWindow()
  window.setMinimumSize(1024,800)
  window.show()
  qtbot.addWidget(window)

  path = qtbot.screenshot(window)
  print(path)
  # saved in
  #   /tmp/pytest-of-steffen/pytest-0/test_simple0/
  #   /tmp/pytest-of-runner/pytest-0/test_simple0/screenshot_MainWindow.png



""" Plain text ediotr that has the special feature of Alt-Up/Down to move text """
from PySide6.QtCore import Qt
from PySide6.QtGui import QKeyEvent, QTextCursor
from PySide6.QtWidgets import QPlainTextEdit


class TextEditor(QPlainTextEdit):
  """ Plain text ediotr that has the special feature of Alt-Up/Down to move text """
  def __init__(self, text:str):
    super().__init__(text)
    self.setTabStopDistance(20)


  def keyPressEvent(self, event: QKeyEvent) -> None:
    """
    Captures key press events. If Alt+Up is pressed, move selected lines up.
    Otherwise, call the default behavior.
    """
    if event.key() == Qt.Key_Up and event.modifiers() == Qt.AltModifier:          # type: ignore[attr-defined]
      self.move_selected_lines(True)
    elif event.key() == Qt.Key_Down and event.modifiers() == Qt.AltModifier:      # type: ignore[attr-defined]
      self.move_selected_lines(False)
    else:
      super().keyPressEvent(event)


  def move_selected_lines(self, moveUp:bool) -> None:
    """Moves the selected lines up by swapping them with the previous line."""
    # define cursor; initial start and end position
    cursor = self.textCursor()  # Get the current text cursor
    if not cursor.hasSelection():  # Expand selection to full lines if there is no selection
      cursor.movePosition(QTextCursor.StartOfBlock, QTextCursor.KeepAnchor)       # type: ignore[attr-defined]
    start = cursor.selectionStart()
    end = cursor.selectionEnd()

    # Move the cursor to the start of the first selected line
    cursor.setPosition(start)
    cursor.movePosition(QTextCursor.StartOfBlock)                                 # type: ignore[attr-defined]
    start_block_pos = cursor.position()  # Store the start of the first selected line

    # Move the cursor to the end of the last selected line
    cursor.setPosition(end)
    if cursor.atBlockStart() and start != end:  # Adjust selection if it ends at a new block
      cursor.movePosition(QTextCursor.PreviousBlock)                              # type: ignore[attr-defined]
    cursor.movePosition(QTextCursor.EndOfBlock)                                   # type: ignore[attr-defined]
    end_block_pos = cursor.position()  # Store the end of the last selected line

    # Select the entire block of selected lines
    cursor.setPosition(start_block_pos)
    cursor.setPosition(end_block_pos+1, QTextCursor.KeepAnchor)                  # type: ignore[attr-defined]
    selected_text = cursor.selection().toPlainText()  # Get the selected text
    cursor.removeSelectedText()

    # Move to the previous / next block (line above/below the selected text)
    cursor.setPosition(start_block_pos)
    if cursor.atStart():  # If we are at the very first line, we cannot move up
      return
    cursor.movePosition(QTextCursor.PreviousBlock if moveUp else QTextCursor.NextBlock)# type: ignore[attr-defined]
    cursor.insertText(selected_text)  # Replace the previous line with selected text
    return



#!/usr/bin/python3
"""TEST using the FULL set of python-requirements: create the default example that all installations create and verify it thoroughly """
import logging, warnings, unittest, tempfile, os
from pathlib import Path
from zipfile import ZipFile
from pasta_eln.backend import Backend
from pasta_eln.inputOutput import exportELN

class TestStringMethods(unittest.TestCase):
  """
  derived class for this test
  """
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.be = None
    self.dirName = ''

  def test_main(self):
    """
    main function
    """
    # initialization: create database, destroy on filesystem and database and then create new one
    warnings.filterwarnings('ignore', message='numpy.ufunc size changed')
    warnings.filterwarnings('ignore', message='invalid escape sequence')
    warnings.filterwarnings('ignore', category=ResourceWarning, module='PIL')
    warnings.filterwarnings('ignore', category=ImportWarning)
    logPath = Path.home()/'pastaELN.log'
    logging.basicConfig(filename=logPath, level=logging.INFO, format='%(asctime)s|%(levelname)s:%(message)s',
                        datefmt='%m-%d %H:%M:%S')   #This logging is always info, since for installation only
    for package in ['urllib3', 'requests', 'asyncio', 'PIL', 'matplotlib.font_manager']:
      logging.getLogger(package).setLevel(logging.WARNING)
    logging.info('Start test')

    # create .eln
    self.be = Backend('research')
    projID = self.be.output('x0').split('|')[-1].strip()
    tempDir = tempfile.gettempdir()
    fileName = f'{tempDir}/PASTA.eln'
    allDocTypes = self.be.db.dataHierarchy('','')+['-']
    print(f'Filename {fileName}: docTypes: {", ".join(allDocTypes)}')
    exportELN(self.be, [projID], fileName, allDocTypes)

    # test file
    with ZipFile(fileName, 'r') as zObject:
      zObject.extractall(path=tempDir)
    os.system(f'tree {tempDir}/PASTA')
    print("""
Data on disk:
├── pastaELN.db
├── PastasExampleProject
│   ├── 000_ThisIsAnExampleTask
│   ├── 001_ThisIsAnotherExampleTask
│   │   ├── 000_ThisIsAnExampleSubtask
│   │   ├── 001_ThisIsAnotherExampleSubtask
│   │   └── simple.png
│   └── 002_DataFiles
│       ├── simple.csv
│       ├── simple.png
│       └── story.odt
└── StandardOperatingProcedures
    └── Example_SOP.md
""")

    # try:
    #   os.system(f'code {tempDir}/PASTA/ro-crate-metadata.json')
    # except Exception:
    #   pass
    return


  def tearDown(self):
    logging.info('End test')
    return

if __name__ == '__main__':
  unittest.main()



""" dialog to edit docType schema """
from enum import Enum
from typing import Any
import numpy as np
import pandas as pd
from PySide6.QtWidgets import (QComboBox, QDialog, QDialogButtonBox, QInputDialog,  # pylint: disable=no-name-in-module
                               QMessageBox, QTabBar, QTableWidget, QTableWidgetItem, QTabWidget, QVBoxLayout)
from ...guiCommunicate import Communicate
from ...guiStyle import IconButton, Label, TextButton, showMessage, widgetAndLayout
from .delete_column_delegate import DeleteColumnDelegate
from .docTypeEdit import DocTypeEditor
from .listFreeDelegate import ListFreeDelegate
from .listItemDelegate import ListItemDelegate
from .mandatory_column_delegate import MandatoryColumnDelegate
from .name_column_delegate import NameColumnDelegate
from .reorder_column_delegate import ReorderColumnDelegate

#                0       1            2      3           4      5         6
COLUMN_NAMES = ['name','description','unit','mandatory','item list','free list','move up','delete']
COLUMN_WIDTH = [100,   250,          60,    80,         150,        200,        60,       50]
pd.options.mode.copy_on_write = True

class SchemeEditor(QDialog):
  """ dialog to edit docType schema """
  def __init__(self, comm:Communicate):
    """
    Initialization

    Args:
      comm (Communicate): communication channel
      doc (dict):  document to change / create
    """
    super().__init__()
    self.comm = comm
    self.db   = self.comm.backend.db
    self.docType = ''
    self.docLabel= ''
    self.closeButtons:list[IconButton] = []  #close buttons of tabs
    self.setWindowTitle('Data scheme editor')

    # GUI elements
    self.setMinimumWidth(int(np.sum(COLUMN_WIDTH))+80)
    self.setMinimumHeight(500)
    mainL = QVBoxLayout(self)
    Label('Schema-editor for document types', 'h1', mainL)
    Label('Warning: basic verification exists only. Use with care.', 'h2', mainL)
    _, docTypeL = widgetAndLayout('H', mainL, 's')
    self.selectDocType = QComboBox()
    self.selectDocType.currentTextChanged.connect(self.changeDocType)
    self.selectDocType.setStyleSheet(self.comm.palette.get('secondaryText','color'))
    docTypeL.addWidget(self.selectDocType)
    docTypeL.addStretch(1)
    IconButton('fa5s.plus',  self, [Command.NEW],  docTypeL, tooltip='New document type')
    IconButton('fa5s.edit',  self, [Command.EDIT], docTypeL, tooltip='Edit document type')
    IconButton('fa5s.trash', self, [Command.DEL],  docTypeL, tooltip='Delete document type')

    # tabs: empty
    self.tabW = QTabWidget(self)
    self.tabW.tabBarDoubleClicked.connect(self.renameTab)
    self.tabW.tabBarClicked.connect(self.createNewTab)
    mainL.addWidget(self.tabW, stretch=10)
    self.nameColumnDelegates     :list[NameColumnDelegate]      = []
    self.mandatoryColumnDelegates:list[MandatoryColumnDelegate] = []
    self.listItemDelegates       :list[ListItemDelegate]        = []
    self.listFreeDelegates       :list[ListFreeDelegate]        = []
    self.reorderColumnDelegates  :list[ReorderColumnDelegate]   = []
    self.deleteColumnDelegates   :list[DeleteColumnDelegate]    = []
    self.newWidget = QTableWidget()
    #final button box
    mainL.addStretch(1)
    buttonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Save | QDialogButtonBox.StandardButton.Cancel)
    buttonBox.clicked.connect(self.closeDialog)
    mainL.addWidget(buttonBox)
    #initialize
    _ = [self.selectDocType.addItem(text, data) for (data, text) in self.db.dataHierarchy('','title')]
    self.selectDocType.setCurrentIndex(0)
    return


  def changeDocType(self, label:str) -> None:
    """ Change the document type

    Args:
      label (str): label of document type
    """
    # get data frame
    if not label:
      return
    if self.docType:
      self.finishDocType()
    self.docLabel= label
    self.docType = [data for (data, text) in self.db.dataHierarchy('','title') if text==label][0]
    cmd = 'SELECT docTypeSchema.docType, docTypeSchema.class, docTypeSchema.idx, docTypeSchema.name, '\
          'docTypeSchema.unit, docTypeSchema.mandatory, docTypeSchema.list, definitions.long '\
          'FROM docTypeSchema LEFT JOIN definitions ON definitions.key = (docTypeSchema.class || "." || docTypeSchema.name) '\
          f'WHERE docTypeSchema.docType=="{self.docType}"'
    df = pd.read_sql_query(cmd, self.db.connection).fillna('')
    df.rename(columns={'long':'description'}, inplace=True)
    df['item list'] = df['list'].apply(lambda x: '' if ',' in x else x)
    df['free list'] = df['list'].apply(lambda x: x  if ',' in x else '')
    df = df.drop('list', axis=1)
    df['idx'] = df['idx'].astype('int')
    # GUI
    self.tabW.clear()
    self.reorderColumnDelegates.clear()
    for group in df['class'].unique():
      data  = df[df['class']==group]
      table = QTableWidget(len(data)+1, 8)
      table.verticalHeader().hide()
      table.setAlternatingRowColors(True)
      table.setHorizontalHeaderLabels(COLUMN_NAMES)
      table.horizontalHeader().setStyleSheet('QHeaderView::section {padding: 1px; margin: 0px;}')
      for idx, width in enumerate(COLUMN_WIDTH[:-1]):
        table.setColumnWidth(idx, width)
      for _,row in data.iterrows():
        idx = int(row['idx'])
        for col in range(6):
          table.setItem(idx, col, QTableWidgetItem(row[COLUMN_NAMES[col]]))
      #Delegates to give function to row
      self.nameColumnDelegates.append(NameColumnDelegate())
      table.setItemDelegateForColumn(0, self.nameColumnDelegates[-1])
      self.mandatoryColumnDelegates.append(MandatoryColumnDelegate())
      table.setItemDelegateForColumn(3, self.mandatoryColumnDelegates[-1])
      self.listItemDelegates.append(ListItemDelegate(self.db.dataHierarchy('','title')))
      table.setItemDelegateForColumn(4, self.listItemDelegates[-1])
      self.listFreeDelegates.append(ListFreeDelegate())
      table.setItemDelegateForColumn(5, self.listFreeDelegates[-1])
      self.reorderColumnDelegates.append(ReorderColumnDelegate())
      table.setItemDelegateForColumn(6, self.reorderColumnDelegates[-1])
      self.deleteColumnDelegates.append(DeleteColumnDelegate())
      table.setItemDelegateForColumn(7, self.deleteColumnDelegates[-1])
      self.tabW.addTab(table, group or 'default')
    # define close buttons on some of the tabs
    self.closeButtons.clear()
    for idx in range(1, self.tabW.count()):
      self.closeButtons.append(IconButton('fa5s.times', self, [Command.DEL_GROUP,idx], None, 'Delete group'))
      self.tabW.tabBar().setTabButton(idx, QTabBar.ButtonPosition.RightSide, self.closeButtons[-1])
      header = table.horizontalHeader()
      header.setStretchLastSection(True)
    self.tabW.addTab(self.newWidget, '+')
    return


  def finishDocType(self) -> None:
    """
    Finish a docType and save to DB,
    - either by selecting another docType or by pressing save
    """
    df = self.table2schema()
    # verification: uniqueness in names. etc.
    unique =df['name'].nunique()==df.shape[0]
    if not unique:
      showMessage(self, 'Error', 'Within each table, the text in the first column has to be unique. E.g. no '
                  'two "tags" are allowed.', 'Critical')
      return
    # SAVE DATA
    dfSchema = df.drop('description', axis=1)
    dfDef = df[['class','name','description']]
    dfDef['key'] = dfDef['class']+'.'+dfDef['name']
    dfDef = dfDef.drop(['name','class'], axis=1)[['key','description']]
    dfDef.rename(columns={'description':'long'}, inplace=True)
    nonUnique = dict(dfDef.groupby(['key']).apply(lambda x: len(np.unique(x))))
    if nonUniqueStr:= ', '.join(k for k,v in nonUnique.items() if v>2):
      button = QMessageBox.question(self, 'Non unique definitions',
                                  f'The definitions are non-unique for {nonUniqueStr}. Do you want to flatten?',
                                  QMessageBox.StandardButton.No, QMessageBox.StandardButton.Yes)
      if button == QMessageBox.StandardButton.No:
        return
    dfDef = dfDef.groupby(['key']).first()
    listDef = list(dfDef.itertuples(name=None))
    # save to sqlite
    cmd = 'INSERT INTO definitions (key, long, PURL) VALUES (?, ?, "") ON CONFLICT(key) DO '\
          'UPDATE SET long = excluded.long;'
    self.db.cursor.executemany(cmd, listDef)
    self.db.cursor.execute(f"DELETE FROM docTypeSchema WHERE docType == '{self.docType}'")
    self.db.connection.commit()
    dfSchema.to_sql('docTypeSchema', self.db.connection, if_exists='append', index=False, dtype='str')
    return


  def closeDialog(self, btn:TextButton) -> None:
    """
    cancel or save entered data

    Args:
      btn (QButton): save or cancel button
    """
    if btn.text().endswith('Cancel'):
      self.reject()
    elif btn.text().endswith('Save'):
      self.finishDocType()
      self.accept()
    return


  def execute(self, command:list[Any]) -> None:
    """
    Event if user clicks button in the center

    Args:
      command (list): list of commands
    """
    if command[0] is Command.NEW:
      dialog = DocTypeEditor(self.comm, '', self.changeDocType)
      dialog.exec()
      docLabel = str(self.docLabel)
      self.selectDocType.clear()
      _ = [self.selectDocType.addItem(text, data) for (data, text) in self.db.dataHierarchy('','title')]
      self.selectDocType.setCurrentText(docLabel)
    elif command[0] is Command.EDIT:
      dialog = DocTypeEditor(self.comm, self.selectDocType.currentData())
      dialog.exec()
      docIdx = self.selectDocType.currentIndex()
      self.selectDocType.clear()
      _ = [self.selectDocType.addItem(text, data) for (data, text) in self.db.dataHierarchy('','title')]
      self.selectDocType.setCurrentIndex(docIdx)
    elif command[0] is Command.DEL:
      button = QMessageBox.question(self, 'Question', 'Do you really want to remove the doc-type?',
                                    QMessageBox.StandardButton.No, QMessageBox.StandardButton.Yes)
      if button == QMessageBox.StandardButton.No:
        return
      self.db.cursor.execute(f"DELETE FROM docTypes WHERE docType == '{self.selectDocType.currentData()}'")
      self.db.cursor.execute(f"DELETE FROM docTypeSchema WHERE docType == '{self.selectDocType.currentData()}'")
      self.db.connection.commit()
      self.selectDocType.clear()
      _ = [self.selectDocType.addItem(text, data) for (data, text) in self.db.dataHierarchy('','title')]
    elif command[0] is Command.DEL_GROUP:
      docLabel = str(self.docLabel)
      button = QMessageBox.question(self, 'Question', 'Do you really want to remove this group?',
                                    QMessageBox.StandardButton.No, QMessageBox.StandardButton.Yes)
      if button == QMessageBox.StandardButton.Yes:
        group = self.tabW.tabBar().tabText(command[1])
        self.db.cursor.execute(f"DELETE FROM docTypeSchema WHERE docType == '{self.docType}' "\
                                            f"AND class == '{group}'")
        self.db.connection.commit()
        self.changeDocType(docLabel)
    return


  def table2schema(self) -> pd.DataFrame:
    """ save text content of tables """
    df = pd.DataFrame()
    lastRow = 0
    for tabID in range(self.tabW.count()-1):  # -1: ignore last tab +
      group = self.tabW.tabBar().tabText(tabID)
      group = '' if group=='default' else group
      widget:QTableWidget = self.tabW.widget(tabID) # type: ignore[assignment]
      for row in range(widget.rowCount()-1):  # -1: ignore last row as it is meant for adding
        df.at[lastRow, 'docType'] = self.docType
        df.at[lastRow, 'class']   = group
        df.at[lastRow, 'idx']     = str(row)
        for col in range(6):
          item = widget.item(row,col)
          df.at[lastRow, COLUMN_NAMES[col]] = '' if item is None else item.text()
        lastRow += 1
    df['list'] = df['item list']+df['free list']
    df.drop(['item list','free list'], axis=1, inplace=True)
    return df


  def renameTab(self, idx:int) -> None:
    """ rename tab after double clicking

    Args:
      idx (int): index of the tab
    """
    if 0 < idx <self.tabW.count()-1:
      textOld = self.tabW.tabText(idx)
      textNew, ok = QInputDialog.getText(self, 'Rename group', 'Enter new group name:', text=textOld)
      if ok and textNew.strip():
        self.tabW.setTabText(idx, textNew.strip())
    return


  def createNewTab(self, idx:int) -> None:
    """ Create new group

    Args:
      idx (int): index of the tab
    """
    if idx == self.tabW.count()-1:
      textNew, ok = QInputDialog.getText(self, 'Create group', 'Enter new group name:', text='')
      if ok and textNew.strip():
        self.db.cursor.execute(f"INSERT INTO docTypeSchema VALUES ({', '.join(['?']*7)})",
                        [self.docType, textNew.strip(), '0', 'item', '', '', ''])
        self.db.connection.commit()
        self.changeDocType(self.docLabel)
    return


class Command(Enum):
  """ Commands used in this file """
  NEW          = 1
  DEL          = 2
  EDIT         = 3
  DEL_GROUP    = 4
